#!/usr/bin/python
# -*- coding: ascii -*-
# $Id$
#
# Author:	Filippo Rivato
# Date: 16 October 2015

__author__ = "Filippo Rivato"
__email__  = "f.rivato@gmail.com"

__name__ = _("Pyrograph")
__version__= "0.0.3"

from CNC import CNC,Block
from ToolsPage import Plugin

#==============================================================================
#Pyrograph class
#==============================================================================
class Pyrograph:
	def __init__(self,name="Pyrograph"):
		self.name = name


#==============================================================================
# Create pyrograph
#==============================================================================
class Tool(Plugin):
	__doc__ = _("Create a variable feed path based upon image brightness")
	def __init__(self, master):
		Plugin.__init__(self, master, "Pyrograph")
		self.icon  = "pyrograph"
		self.group = "Artistic"

		self.variables = [
			("name",         "db" ,    "", _("Name")),
			("ToolSize",     "mm" ,    0.5, _("Pyrograph tip size")),
			("Depth",        "mm" ,    0.0, _("Working Depth")),
			("MaxSize",      "mm" ,  100.0, _("Maximum size")),
			("FeedMin",     "int" ,    250, _("Minimum feed")),
			("FeedMax",     "int" ,   5000, _("Maximum feed")),
			("Direction", "Horizontal,Vertical,Both", "Horizontal", _("Direction")),
			("DrawBorder",  "bool",  False, _("Draw border")),
			("File",       "file" ,     "", _("Image to process")),
		]
		self.buttons.append("exe")

	# ----------------------------------------------------------------------
	def execute(self, app):
		try:
			from PIL import Image
		except:
			app.setStatus(_("Pyrograph abort: This plugin requires PIL/Pillow"))
			return

		n = self["name"]
		if not n or n=="default": n="Pyrograph"

		#Calc desired size
		toolSize = self.fromMm("ToolSize")
		maxSize = self.fromMm("MaxSize")
		feedMin = self["FeedMin"]
		feedMax = self["FeedMax"]
		depth = self.fromMm("Depth")
		direction = self["Direction"]
		drawBorder = self["DrawBorder"]

		#Check parameters
		if direction is "":
			app.setStatus(_("Pyrograph abort: please define a scan Direction"))
			return

		if toolSize <=0:
			app.setStatus(_("Pyrograph abort: Tool Size must be > 0"))
			return

		if feedMin <=0 or feedMax <=0 :
			app.setStatus(_("Pyrograph abort: Please check feed rate parameters"))
			return

		#divisions
		divisions = maxSize / toolSize

		fileName = self["File"]
		try:
			img = Image.open(fileName)
			img = img.convert ('RGB') #be sure to have color to calculate luminance
		except:
			app.setStatus(_("Pyrograph abort: Can't read image file"))
			return

		iWidth,iHeight =  img.size
		newWidth = iWidth
		newHeight = iHeight

		ratio = 1
		if (iWidth > iHeight):
			ratio = float(iWidth) / float(iHeight)
			newWidth = int(divisions)
			newHeight = int(divisions / ratio)
		else:
			ratio = float(iHeight) / float(iWidth)
			newWidth = int(divisions / ratio)
			newHeight = int(divisions)

		#Create a thumbnail image to work faster
		img.thumbnail((newWidth,newHeight), Image.ANTIALIAS)
		newWidth,newHeight =  img.size
		#img.save("thumb.png")
		pixels = list(img.getdata())

		#Extract luminance
		gMap = []
		for x in range(0,newWidth):
			gRow = []
			for y in range(0,newHeight):
				R,G,B = pixels[(y * newWidth) + x ]
				L = (0.299*R + 0.587*G + 0.114*B) #Luminance (Rec. 601 standard)
				gRow.append(L)
			gMap.append(gRow)

		#Init blocks
		blocks = []
		block = Block(self.name)
		block.append("(Pyrograph W=%g x H=%g x D=%g)" %
		(newWidth * toolSize , newHeight * toolSize , depth))

		#Create points for vertical scan
		xH = []
		yH = []
		fH = []
		if (direction=="Vertical" or direction=="Both"):
			r = range(0,newHeight)
			for x in range(0,newWidth):
				r = r[::-1]
				fPrec = -1
				for y in r:
					f = int(feedMin + ((feedMax - feedMin) * gMap[x][y] / 255.0))
					if(f != fPrec or y==0 or  y==newHeight-1):
						xH.append(x * toolSize)
						yH.append((newHeight-y) * toolSize)
						fH.append(f)
					fPrec = f

		#Create points for horizontal scan
		xV = []
		yV = []
		fV = []
		if (direction=="Horizontal" or direction=="Both"):
			r = range(0,newWidth)
			for y in reversed(range(0,newHeight)):
				fPrec = -1
				for x in r:
					f = int(feedMin + ((feedMax - feedMin) * gMap[x][y] / 255.0))
					if(f != fPrec or x==0 or x==newWidth-1):
						xV.append(x * toolSize)
						yV.append((newHeight-y) * toolSize)
						fV.append(f)
					fPrec = f
				r = r[::-1]

		#Gcode Horizontal
		if (len(xH)>1 and len(yH)>1):
			block.append(CNC.zsafe())
			block.append(CNC.grapid(xH[0],yH[0]))
			block.append(CNC.zenter(depth))
			for x,y,f in zip(xH,yH,fH):
					v = (x,y,depth)
					block.append(CNC.glinev(1,v,f))

		#Gcode Vertical
		if (len(xV)>1 and len(yV)>1):
			block.append(CNC.zsafe())
			block.append(CNC.grapid(xV[0],yV[0]))
			block.append(CNC.zenter(depth))
			for x,y,f in zip(xV,yV,fV):
					v = (x,y,depth)
					block.append(CNC.glinev(1,v,f))

		#Draw Border if required
		if drawBorder:
			block.append(CNC.zsafe())
			block.append(CNC.grapid(0,0))
			block.append(CNC.zenter(depth))
			block.append(CNC.gcode(1, [("f",feedMin)]))
			block.append(CNC.gline(newWidth * toolSize - toolSize,0))
			block.append(CNC.gline(newWidth * toolSize - toolSize ,newHeight* toolSize))
			block.append(CNC.gline(0,newHeight* toolSize ))
			block.append(CNC.gline(0,0))

		#Gcode Zsafe
		block.append(CNC.zsafe())

		blocks.append(block)
		active = app.activeBlock()
		if active==0: active=1
		app.gcode.insBlocks(active, blocks, "Pyrograph")
		app.refresh()
		app.setStatus(_("Generated Pyrograph W=%g x H=%g x D=%g") %
		(newWidth * toolSize , newHeight * toolSize , depth))



#!/usr/bin/python
# -*- coding: ascii -*-
# $Id: bCNC.py,v 1.6 2014/10/15 15:04:48 bnv Exp bnv $
#
# Author: vvlachoudis@gmail.com
# Date: 17-Jun-2015

__author__  = "Vasilis Vlachoudis"
__email__   = "vvlachoudis@gmail.com"

import os
import re
import sys
import glob
import traceback
import rexx
import time
import threading
import webbrowser

from datetime import datetime
try:
	import serial
except:
	serial = None
try:
	from Queue import *
except ImportError:
	from queue import *

from CNC import WAIT, MSG, UPDATE, WCS, CNC, GCode
import Utils
import Pendant

WIKI = "https://github.com/vlachoudis/bCNC/wiki"

SERIAL_POLL    = 0.125	# s
SERIAL_TIMEOUT = 0.10	# s
G_POLL	       = 10	# s
RX_BUFFER_SIZE = 128

OV_FEED_100     = chr(0x90)        # Extended override commands
OV_FEED_i10     = chr(0x91)
OV_FEED_d10     = chr(0x92)
OV_FEED_i1      = chr(0x93)
OV_FEED_d1      = chr(0x94)

OV_RAPID_100    = chr(0x95)
OV_RAPID_50     = chr(0x96)
OV_RAPID_25     = chr(0x97)

OV_SPINDLE_100  = chr(0x99)
OV_SPINDLE_i10  = chr(0x9A)
OV_SPINDLE_d10  = chr(0x9B)
OV_SPINDLE_i1   = chr(0x9C)
OV_SPINDLE_d1   = chr(0x9D)

OV_SPINDLE_STOP = chr(0x9E)

OV_FLOOD_TOGGLE = chr(0xA0)
OV_MIST_TOGGLE  = chr(0xA1)

GPAT	  = re.compile(r"[A-Za-z]\s*[-+]?\d+.*")
STATUSPAT = re.compile(r"^<(\w*?),MPos:([+\-]?\d*\.\d*),([+\-]?\d*\.\d*),([+\-]?\d*\.\d*),WPos:([+\-]?\d*\.\d*),([+\-]?\d*\.\d*),([+\-]?\d*\.\d*),?(.*)>$")
POSPAT	  = re.compile(r"^\[(...):([+\-]?\d*\.\d*),([+\-]?\d*\.\d*),([+\-]?\d*\.\d*):?(\d*)\]$")
TLOPAT	  = re.compile(r"^\[(...):([+\-]?\d*\.\d*)\]$")
DOLLARPAT = re.compile(r"^\[G\d* .*\]$")
FEEDPAT   = re.compile(r"^(.*)[fF](\d+\.?\d+)(.*)$")
SPLITPAT  = re.compile(r"[:,]")
VARPAT    = re.compile(r"^\$(\d+)=(\d*\.?\d*) *\(?.*")

CONNECTED     = "Connected"
NOT_CONNECTED = "Not connected"

STATECOLOR = {	"Alarm"       : "Red",
		"Run"	      : "LightGreen",
		"Hold"	      : "Orange",
		"Hold:0"      : "Orange",
		"Hold:1"      : "Orange",
		CONNECTED     : "Orange",
		NOT_CONNECTED : "OrangeRed"
		}

STATE_DESC = {
		"Hold:0" : "Hold complete",
		"Hold:1" : "Hold in-progress",
		"Door:0" : "Door closed",
		"Door:1" : "Machine stopped",
		"Door:2" : "Door opened",
		"Door:3" : "Door closed and resuming",
	}

STATECOLORDEF = "LightYellow"

# From https://github.com/grbl/grbl/wiki/Interfacing-with-Grbl
ERROR_CODES = {
	"Run"	   : _("bCNC is currently sending a gcode program to Grbl"),
	"Idle"	   : _("Grbl is in idle state and waiting for user commands"),
	"Hold"	   : _("Grbl is on hold state. Click on resume (pause) to continue"),
	"Alarm"    : _("Alarm is an emergency state. Something has gone terribly wrong when these occur. Typically, they are caused by limit error when the machine has moved or wants to move outside the machine space and crash into something. They also report problems if Grbl is lost and can't guarantee positioning or a probe command has failed. Once in alarm-mode, Grbl will lock out and shut down everything until the user issues a reset. Even after a reset, Grbl will remain in alarm-mode, block all G-code from being executed, but allows the user to override the alarm manually. This is to ensure the user knows and acknowledges the problem and has taken steps to fix or account for it."),
	NOT_CONNECTED : _("Grbl is not connected. Please specify the correct port and click Open."),
	CONNECTED     : _("Connection is established with Grbl"),

	"ok" : _("All is good! Everything in the last line was understood by Grbl and was successfully processed and executed."),

	"error:1"  : _("G-code words consist of a letter and a value. Letter was not found."),
	"error:2"  : _("Numeric value format is not valid or missing an expected value."),
	"error:3"  : _("Grbl '$' system command was not recognized or supported."),
	"error:4"  : _("Negative value received for an expected positive value."),
	"error:5"  : _("Homing cycle is not enabled via settings."),
	"error:6"  : _("Minimum step pulse time must be greater than 3usec"),
	"error:7"  : _("EEPROM read failed. Reset and restored to default values."),
	"error:8"  : _("Grbl '$' command cannot be used unless Grbl is IDLE. Ensures smooth operation during a job."),
	"error:9"  : _("G-code locked out during alarm or jog state"),
	"error:10" : _("Soft limits cannot be enabled without homing also enabled."),
	"error:11" : _("Max characters per line exceeded. Line was not processed and executed."),
	"error:12" : _("(Compile Option) Grbl '$' setting value exceeds the maximum step rate supported."),
	"error:13" : _("Safety door detected as opened and door state initiated."),
	"error:14" : _("(Grbl-Mega Only) Build info or startup line exceeded EEPROM line length limit."),
	"error:15" : _("Jog target exceeds machine travel. Command ignored."),
	"error:16" : _("Jog command with no '=' or contains prohibited g-code."),
	"error:17" : _("Laser mode requires PWM output."),
	"error:20" : _("Unsupported or invalid g-code command found in block."),
	"error:21" : _("More than one g-code command from same modal group found in block."),
	"error:22" : _("Feed rate has not yet been set or is undefined."),
	"error:23" : _("G-code command in block requires an integer value."),
	"error:24" : _("Two G-code commands that both require the use of the XYZ axis words were detected in the block."),
	"error:25" : _("A G-code word was repeated in the block."),
	"error:26" : _("A G-code command implicitly or explicitly requires XYZ axis words in the block, but none were detected."),
	"error:27" : _("N line number value is not within the valid range of 1 - 9,999,999."),
	"error:28" : _("A G-code command was sent, but is missing some required P or L value words in the line."),
	"error:29" : _("Grbl supports six work coordinate systems G54-G59. G59.1, G59.2, and G59.3 are not supported."),
	"error:30" : _("The G53 G-code command requires either a G0 seek or G1 feed motion mode to be active. A different motion was active."),
	"error:31" : _("There are unused axis words in the block and G80 motion mode cancel is active."),
	"error:32" : _("A G2 or G3 arc was commanded but there are no XYZ axis words in the selected plane to trace the arc."),
	"error:33" : _("The motion command has an invalid target. G2, G3, and G38.2 generates this error, if the arc is impossible to generate or if the probe target is the current position."),
	"error:34" : _("A G2 or G3 arc, traced with the radius definition, had a mathematical error when computing the arc geometry. Try either breaking up the arc into semi-circles or quadrants, or redefine them with the arc offset definition."),
	"error:35" : _("A G2 or G3 arc, traced with the offset definition, is missing the IJK offset word in the selected plane to trace the arc."),
	"error:36" : _("There are unused, leftover G-code words that aren't used by any command in the block."),
	"error:37" : _("The G43.1 dynamic tool length offset command cannot apply an offset to an axis other than its configured axis. The Grbl default axis is the Z-axis."),

	"ALARM:1" : _("Hard limit triggered. Machine position is likely lost due to sudden and immediate halt. Re-homing is highly recommended."),
	"ALARM:2" : _("G-code motion target exceeds machine travel. Machine position safely retained. Alarm may be unlocked."),
	"ALARM:3" : _("Reset while in motion. Grbl cannot guarantee position. Lost steps are likely. Re-homing is highly recommended."),
	"ALARM:4" : _("Probe fail. The probe is not in the expected initial state before starting probe cycle, where G38.2 and G38.3 is not triggered and G38.4 and G38.5 is triggered."),
	"ALARM:5" : _("Probe fail. Probe did not contact the workpiece within the programmed travel for G38.2 and G38.4."),
	"ALARM:6" : _("Homing fail. Reset during active homing cycle."),
	"ALARM:7" : _("Homing fail. Safety door was opened during active homing cycle."),
	"ALARM:8" : _("Homing fail. Cycle failed to clear limit switch when pulling off. Try increasing pull-off setting or check wiring."),
	"ALARM:9" : _("Homing fail. Could not find limit switch within search distance. Defined as 1.5 * max_travel on search and 5 * pulloff on locate phases."),

	"Hold:0" : _("Hold complete. Ready to resume."),
	"Hold:1" : _("Hold in-progress. Reset will throw an alarm."),
	"Door:0" : _("Door closed. Ready to resume."),
	"Door:1" : _("Machine stopped. Door still ajar. Can't resume until closed."),
	"Door:2" : _("Door opened. Hold (or parking retract) in-progress. Reset will throw an alarm."),
	"Door:3" : _("Door closed and resuming. Restoring from park, if applicable. Reset will throw an alarm."),
}


# Convert Grbl V1.0 codes to Grbl V0.9
for e1,e0 in (	("error: Expected command letter", "error:1"),
		("error: Bad number format", "error:2"),
		("error: Invalid statement", "error:3"),
		("error: Value < 0", "error:4"),
		("error: Setting disabled", "error:5"),
		("error: Value < 3 usec", "error:6"),
		("error: EEPROM read fail. Using defaults", "error:7"),
		("error: Not idle", "error:8"),
		("error: G-code lock", "error:9"),
		("error: Homing not enabled", "error:10"),
		("error: Line overflow", "error:11"),
		("error: Step rate > 30kHz*", "error:12"),
		("error: Check Door", "error:13"),
		("error: Line length exceeded", "error:14"),
		("error: Travel exceeded", "error:15"),
		("error: Invalid jog command", "error:16"),
		("error: Unsupported command", "error:20"),
		("error: Modal group violation", "error:21"),
		("error: Undefined feed rate", "error:22"),
		("error: Invalid gcode ID:23", "error:23"),
		("error: Invalid gcode ID:24", "error:24"),
		("error: Invalid gcode ID:25", "error:25"),
		("error: Invalid gcode ID:26", "error:26"),
		("error: Invalid gcode ID:27", "error:27"),
		("error: Invalid gcode ID:28", "error:28"),
		("error: Invalid gcode ID:29", "error:29"),
		("error: Invalid gcode ID:30", "error:30"),
		("error: Invalid gcode ID:31", "error:31"),
		("error: Invalid gcode ID:32", "error:32"),
		("error: Invalid gcode ID:33", "error:33"),
		("error: Invalid gcode ID:34", "error:34"),
		("error: Invalid gcode ID:35", "error:35"),
		("error: Invalid gcode ID:36", "error:36"),
		("error: Invalid gcode ID:37", "error:37"),
		("ALARM: Hard limit", "ALARM:1"),
		("ALARM: Soft limit", "ALARM:2"),
		("ALARM: Abort during cycle", "ALARM:3"),
		("ALARM: Probe fail", "ALARM:4"),
		("ALARM: Probe fail", "ALARM:5"),
		("ALARM: Homing fail", "ALARM:6"),
		("ALARM: Homing fail", "ALARM:7"),
		("ALARM: Homing fail", "ALARM:8"),
		("ALARM: Homing fail", "ALARM:9") ):
	ERROR_CODES[e1] = ERROR_CODES[e0]

#==============================================================================
# bCNC Generic controller class
#==============================================================================
class ControllerGeneric:
	def test(self):
		print("test super")

#==============================================================================
# bCNC Sender class
#==============================================================================
class Sender:
	# Messages types for log Queue
	MSG_BUFFER  =  0	# write to buffer one command
	MSG_SEND    =  1	# send message
	MSG_RECEIVE =  2	# receive message from controller
	MSG_OK	    =  3	# ok response from controller, move top most command to terminal
	MSG_ERROR   =  4	# error message or exception
	MSG_RUNEND  =  5	# run ended
	MSG_CLEAR   =  6	# clear buffer

	def __init__(self):
		# Global variables
		self.history	 = []
		self._historyPos = None

		#self.control     = None
		self.controllers = {}
		self.controllerLoad()
		self.controllerSet("GRBL1")

		CNC.loadConfig(Utils.config)
		self.gcode = GCode()
		self.cnc   = self.gcode.cnc

		self.log	 = Queue()	# Log queue returned from GRBL
		self.queue	 = Queue()	# Command queue to be send to GRBL
		self.pendant	 = Queue()	# Command queue to be executed from Pendant
		self.serial	 = None
		self.thread	 = None

		self._posUpdate  = False	# Update position
		self._probeUpdate= False	# Update probe
		self._gUpdate	 = False	# Update $G
		self._update	 = None		# Generic update

		self.running	 = False
		self._runLines	 = 0
		self._quit	 = 0		# Quit counter to exit program
		self._stop	 = False	# Raise to stop current run
		self._pause	 = False	# machine is on Hold
		self._alarm	 = True		# Display alarm message if true
		self._msg	 = None
		self._sumcline	 = 0
		self._lastFeed	 = 0
		self._newFeed	 = 0

		self._onStart    = ""
		self._onStop     = ""


	#----------------------------------------------------------------------
	def controllerLoad(self):
		# Find plugins in the controllers directory and load them
		for f in glob.glob("%s/controllers/*.py"%(Utils.prgpath)):
			name,ext = os.path.splitext(os.path.basename(f))
			#print("Loaded motion controller plugin: %s"%(name))
			try:
				exec("import %s"%(name))
				self.controllers[name] = eval("%s.Controller(self)"%(name))
			except (ImportError, AttributeError):
				typ, val, tb = sys.exc_info()
				traceback.print_exception(typ, val, tb)

	#----------------------------------------------------------------------
	def controllerList(self):
		#print("ctrlist")
		#self.controllers["GRBL1"].test()
		#if len(self.controllers.keys()) < 1: self.controllerLoad()
		return sorted(self.controllers.keys())

	#----------------------------------------------------------------------
	def controllerSet(self, ctl):
		#print("Activating motion controller plugin: %s"%(ctl))
		if ctl in self.controllers.keys():
			self.controller = ctl
			self.control = self.controllers[ctl]
			#self.control.test()


	#----------------------------------------------------------------------
	def quit(self, event=None):
		self.saveConfig()
		Pendant.stop()

	#----------------------------------------------------------------------
	def loadConfig(self):
		self.controllerSet(Utils.getStr("Connection", "controller"))
		Pendant.port	 = Utils.getInt("Connection","pendantport",Pendant.port)
		GCode.LOOP_MERGE = Utils.getBool("File","dxfloopmerge")
		self.loadHistory()

	#----------------------------------------------------------------------
	def saveConfig(self):
		self.saveHistory()

	#----------------------------------------------------------------------
	def loadHistory(self):
		try:
			f = open(Utils.hisFile,"r")
		except:
			return
		self.history = [x.strip() for x in f]
		f.close()

	#----------------------------------------------------------------------
	def saveHistory(self):
		try:
			f = open(Utils.hisFile,"w")
		except:
			return
		f.write("\n".join(self.history))
		f.close()

	#----------------------------------------------------------------------
	# Evaluate a line for possible expressions
	# can return a python exception, needs to be catched
	#----------------------------------------------------------------------
	def evaluate(self, line):
		return self.gcode.evaluate(CNC.compileLine(line,True))

	#----------------------------------------------------------------------
	# Execute a line as gcode if pattern matches
	# @return True on success
	#	  False otherwise
	#----------------------------------------------------------------------
	def executeGcode(self, line):
		if isinstance(line, tuple) or \
		   line[0] in ("$","!","~","?","(","@") or GPAT.match(line):
			self.sendGCode(line)
			return True
		return False

	#----------------------------------------------------------------------
	# Execute a single command
	#----------------------------------------------------------------------
	def executeCommand(self, line):
		#print
		#print "<<<",line
		#try:
		#	line = self.gcode.evaluate(CNC.compileLine(line,True))
		#except:
		#	return "Evaluation error", sys.exc_info()[1]
		#print ">>>",line

		if line is None: return

		oline = line.strip()
		line  = oline.replace(","," ").split()
		cmd   = line[0].upper()

		# ABS*OLUTE: Set absolute coordinates
		if rexx.abbrev("ABSOLUTE",cmd,3):
			self.sendGCode("G90")

		# HELP: open browser to display help
		elif cmd == "HELP":
			self.help()

		# HOME: perform a homing cycle
		elif cmd == "HOME":
			self.home()

		# LO*AD [filename]: load filename containing g-code
		elif rexx.abbrev("LOAD",cmd,2):
			self.load(line[1])

		# OPEN: open serial connection to grbl
		# CLOSE: close serial connection to grbl
		elif cmd in ("OPEN","CLOSE"):
			self.openClose()

		# QU*IT: quit program
		# EX*IT: exit program
		elif rexx.abbrev("QUIT",cmd,2) or rexx.abbrev("EXIT",cmd,2):
			self.quit()

		# PAUSE: pause cycle
		elif cmd == "PAUSE":
			self.pause()

		# RESUME: resume
		elif cmd == "RESUME":
			self.resume()

		# FEEDHOLD: feedhold
		elif cmd == "FEEDHOLD":
			self.feedHold()

		# REL*ATIVE: switch to relative coordinates
		elif rexx.abbrev("RELATIVE",cmd,3):
			self.sendGCode("G91")

		# RESET: perform a soft reset to grbl
		elif cmd == "RESET":
			self.softReset()

		# RUN: run g-code
		elif cmd == "RUN":
			self.run()

		# SAFE [z]: safe z to move
		elif cmd=="SAFE":
			try: CNC.vars["safe"] = float(line[1])
			except: pass
			self.statusbar["text"] = "Safe Z= %g"%(CNC.vars["safe"])

		# SA*VE [filename]: save to filename or to default name
		elif rexx.abbrev("SAVE",cmd,2):
			if len(line)>1:
				self.save(line[1])
			else:
				self.saveAll()

		# SENDHEX: send a hex-char in grbl
		elif cmd == "SENDHEX":
			self.sendHex(line[1])

		# SET [x [y [z]]]: set x,y,z coordinates to current workspace
		elif cmd == "SET":
			try: x = float(line[1])
			except: x = None
			try: y = float(line[2])
			except: y = None
			try: z = float(line[3])
			except: z = None
			self._wcsSet(x,y,z)

		elif cmd == "SET0":
			self._wcsSet(0.,0.,0.)

		elif cmd == "SETX":
			try: x = float(line[1])
			except: x = ""
			self._wcsSet(x,None,None)

		elif cmd == "SETY":
			try: y = float(line[1])
			except: y = ""
			self._wcsSet(None,y,None)

		elif cmd == "SETZ":
			try: z = float(line[1])
			except: z = ""
			self._wcsSet(None,None,z)

		# STOP: stop current run
		elif cmd == "STOP":
			self.stopRun()

		# UNL*OCK: unlock grbl
		elif rexx.abbrev("UNLOCK",cmd,3):
			self.unlock()

		# Send commands to SMOOTHIE
		elif self.control.executeCommand(oline, line, cmd):
			pass

		else:
			return _("unknown command"),_("Invalid command %s")%(oline)

	#----------------------------------------------------------------------
	def help(self, event=None):
		webbrowser.open(WIKI,new=2)

	#----------------------------------------------------------------------
	def loadRecent(self, recent):
		filename = Utils.getRecent(recent)
		if filename is None: return
		self.load(filename)

	#----------------------------------------------------------------------
	def _loadRecent0(self,event): self.loadRecent(0)
	def _loadRecent1(self,event): self.loadRecent(1)
	def _loadRecent2(self,event): self.loadRecent(2)
	def _loadRecent3(self,event): self.loadRecent(3)
	def _loadRecent4(self,event): self.loadRecent(4)
	def _loadRecent5(self,event): self.loadRecent(5)
	def _loadRecent6(self,event): self.loadRecent(6)
	def _loadRecent7(self,event): self.loadRecent(7)
	def _loadRecent8(self,event): self.loadRecent(8)
	def _loadRecent9(self,event): self.loadRecent(9)

	#----------------------------------------------------------------------
	def _saveConfigFile(self, filename=None):
		if filename is None:
			filename = self.gcode.filename
		Utils.setUtf("File", "dir",   os.path.dirname(os.path.abspath(filename)))
		Utils.setUtf("File", "file",  os.path.basename(filename))
		Utils.setUtf("File", "probe", os.path.basename(self.gcode.probe.filename))

	#----------------------------------------------------------------------
	# Load a file into editor
	#----------------------------------------------------------------------
	def load(self, filename):
		fn,ext = os.path.splitext(filename)
		ext = ext.lower()
		if ext==".probe":
			if filename is not None:
				self.gcode.probe.filename = filename
				self._saveConfigFile()
			self.gcode.probe.load(filename)
		elif ext == ".orient":
			# save orientation file
			self.gcode.orient.load(filename)
		elif ext == ".stl" or ext == ".ply":
			# FIXME: implements solid import???
			import tkMessageBox
			tkMessageBox.showinfo("Open 3D Mesh", "Importing of 3D mesh files in .STL and .PLY format is supported by SliceMesh plugin.\nYou can find it in Tools->SliceMesh.")
		elif ext==".dxf":
			self.gcode.init()
			self.gcode.importDXF(filename)
			self._saveConfigFile(filename)
		elif ext==".svg":
			self.gcode.init()
			self.gcode.importSVG(filename)
			self._saveConfigFile(filename)
		else:
			self.gcode.load(filename)
			self._saveConfigFile()
		Utils.addRecent(filename)

	#----------------------------------------------------------------------
	def save(self, filename):
		fn,ext = os.path.splitext(filename)
		ext = ext.lower()
		if ext == ".probe" or ext == ".xyz":
			# save probe
			if not self.gcode.probe.isEmpty():
				self.gcode.probe.save(filename)
			if filename is not None:
				self._saveConfigFile()
		elif ext == ".orient":
			# save orientation file
			self.gcode.orient.save(filename)
		elif ext == ".stl":
			#save probe as STL
			self.gcode.probe.saveAsSTL(filename)
		elif ext == ".dxf":
			return self.gcode.saveDXF(filename)
		elif ext == ".svg":
			return self.gcode.saveSVG(filename)
		elif ext == ".txt":
			#save gcode as txt (only enabled blocks and no bCNC metadata)
			return self.gcode.saveTXT(filename)
		else:
			if filename is not None:
				self.gcode.filename = filename
				self._saveConfigFile()
			Utils.addRecent(self.gcode.filename)
			return self.gcode.save()

	#----------------------------------------------------------------------
	def saveAll(self, event=None):
		if self.gcode.filename:
			self.save(self.gcode.filename)
			if self.gcode.probe.filename:
				self.save(self.gcode.probe.filename)
		return "break"

	#----------------------------------------------------------------------
	# Open serial port
	#----------------------------------------------------------------------
	def open(self, device, baudrate):
		#self.serial = serial.Serial(
		self.serial = serial.serial_for_url(
						device,
						baudrate,
						bytesize=serial.EIGHTBITS,
						parity=serial.PARITY_NONE,
						stopbits=serial.STOPBITS_ONE,
						timeout=SERIAL_TIMEOUT,
						xonxoff=False,
						rtscts=False)
		# Toggle DTR to reset Arduino
		try:
			self.serial.setDTR(0)
		except IOError:
			pass
		time.sleep(1)
		CNC.vars["state"] = CONNECTED
		CNC.vars["color"] = STATECOLOR[CNC.vars["state"]]
		#self.state.config(text=CNC.vars["state"],
		#		background=CNC.vars["color"])
		# toss any data already received, see
		# http://pyserial.sourceforge.net/pyserial_api.html#serial.Serial.flushInput
		self.serial.flushInput()
		try:
			self.serial.setDTR(1)
		except IOError:
			pass
		time.sleep(1)
		self.serial.write(b"\n\n")
		self._gcount = 0
		self._alarm  = True
		self.thread  = threading.Thread(target=self.serialIO)
		self.thread.start()
		return True

	#----------------------------------------------------------------------
	# Close serial port
	#----------------------------------------------------------------------
	def close(self):
		if self.serial is None: return
		try:
			self.stopRun()
		except:
			pass
		self._runLines = 0
		self.thread = None
		time.sleep(1)
		try:
			self.serial.close()
		except:
			pass
		self.serial = None
		CNC.vars["state"] = NOT_CONNECTED
		CNC.vars["color"] = STATECOLOR[CNC.vars["state"]]

	#----------------------------------------------------------------------
	# Send to controller a gcode or command
	# WARNING: it has to be a single line!
	#----------------------------------------------------------------------
	def sendGCode(self, cmd):
		if self.serial and not self.running:
			if isinstance(cmd,tuple):
				self.queue.put(cmd)
			else:
				self.queue.put(cmd+"\n")

	#----------------------------------------------------------------------
	def sendHex(self, hexcode):
		if self.serial is None: return
		self.serial.write(chr(int(hexcode,16)))
		self.serial.flush()

	#----------------------------------------------------------------------
	def hardReset(self):
		self.busy()
		if self.serial is not None:
			self.control.hardResetPre()
			self.openClose()
			self.control.hardResetAfter()
		self.openClose()
		self.stopProbe()
		self._alarm = False
		CNC.vars["_OvChanged"] = True	# force a feed change if any
		self.notBusy()

	#----------------------------------------------------------------------
	def softReset(self, clearAlarm=True):
		if self.serial:
			self.serial.write(b"\030")
		self.stopProbe()
		if clearAlarm: self._alarm = False
		CNC.vars["_OvChanged"] = True	# force a feed change if any

	#----------------------------------------------------------------------
	def unlock(self, clearAlarm=True):
		if clearAlarm: self._alarm = False
		self.sendGCode("$X")

	#----------------------------------------------------------------------
	def home(self, event=None):
		self._alarm = False
		self.sendGCode("$H")

	#----------------------------------------------------------------------
	def viewSettings(self):
		self.control.viewSettings()

	def viewParameters(self):
		self.sendGCode("$#")

	def viewState(self):
		self.sendGCode("$G")

	def viewBuild(self):
		self.control.viewBuild()

	def viewStartup(self):
		self.control.viewStartup()

	def checkGcode(self):
		self.control.checkGcode()

	def grblHelp(self):
		self.control.grblHelp()

	def grblRestoreSettings(self):
		self.control.grblRestoreSettings()

	def grblRestoreWCS(self):
		self.control.grblRestoreWCS()

	def grblRestoreAll(self):
		self.control.grblRestoreAll()

	#----------------------------------------------------------------------
	def goto(self, x=None, y=None, z=None):
		cmd = "G90G0"
		if x is not None: cmd += "X%g"%(x)
		if y is not None: cmd += "Y%g"%(y)
		if z is not None: cmd += "Z%g"%(z)
		self.sendGCode("%s"%(cmd))

	#----------------------------------------------------------------------
	# FIXME Duplicate with ControlPage
	#----------------------------------------------------------------------
	def _wcsSet(self, x, y, z):
		p = WCS.index(CNC.vars["WCS"])
		if p<6:
			cmd = "G10L20P%d"%(p+1)
		elif p==6:
			cmd = "G28.1"
		elif p==7:
			cmd = "G30.1"
		elif p==8:
			cmd = "G92"

		pos = ""
		if x is not None and abs(x)<10000.0: pos += "X"+str(x)
		if y is not None and abs(y)<10000.0: pos += "Y"+str(y)
		if z is not None and abs(z)<10000.0: pos += "Z"+str(z)
		cmd += pos
		self.sendGCode(cmd)
		self.sendGCode("$#")
		self.event_generate("<<Status>>",
			data=(_("Set workspace %s to %s")%(WCS[p],pos)))
			#data=(_("Set workspace %s to %s")%(WCS[p],pos)).encode("utf8"))
		self.event_generate("<<CanvasFocus>>")

	#----------------------------------------------------------------------
	def feedHold(self, event=None):
		if event is not None and not self.acceptKey(True): return
		if self.serial is None: return
		self.serial.write(b"!")
		self.serial.flush()
		self._pause = True

	#----------------------------------------------------------------------
	def resume(self, event=None):
		if event is not None and not self.acceptKey(True): return
		if self.serial is None: return
		self.serial.write(b"~")
		self.serial.flush()
		self._msg   = None
		self._alarm = False
		self._pause = False

	#----------------------------------------------------------------------
	def pause(self, event=None):
		if self.serial is None: return
		if self._pause:
			self.resume()
		else:
			self.feedHold()

	#----------------------------------------------------------------------
	# FIXME ????
	#----------------------------------------------------------------------
	def g28Command(self):
		self.sendGCode("G28.1")

	#----------------------------------------------------------------------
	# FIXME ????
	#----------------------------------------------------------------------
	def g30Command(self):
		self.sendGCode("G30.1")

	#----------------------------------------------------------------------
	def emptyQueue(self):
		while self.queue.qsize()>0:
			try:
				self.queue.get_nowait()
			except Empty:
				break

	#----------------------------------------------------------------------
	def stopProbe(self):
		if self.gcode.probe.start:
			self.gcode.probe.clear()

	#----------------------------------------------------------------------
	def getBufferFill(self):
		return self._sumcline * 100. / RX_BUFFER_SIZE

	#----------------------------------------------------------------------
	def initRun(self):
		self._quit   = 0
		self._pause  = False
		self._paths  = None
		self.running = True
		self.disable()
		self.emptyQueue()
		time.sleep(1)

	#----------------------------------------------------------------------
	# Called when run is finished
	#----------------------------------------------------------------------
	def runEnded(self):
		if self.running:
			self.log.put((Sender.MSG_RUNEND,_("Run ended")))
			self.log.put((Sender.MSG_RUNEND, str(datetime.now())))
			self.log.put((Sender.MSG_RUNEND, str(CNC.vars["msg"])))
			if self._onStop:
				try:
					os.system(self._onStop)
				except:
					pass
		self._runLines = 0
		self._quit     = 0
		self._msg      = None
		self._pause    = False
		self.running   = False
		CNC.vars["running"] = False

	#----------------------------------------------------------------------
	# Purge the buffer of the controller. Unfortunately we have to perform
	# a reset to clear the buffer of the controller
	#---------------------------------------------------------------------
	def purgeController(self):
		self.serial.write(b"!")
		self.serial.flush()
		time.sleep(1)
		# remember and send all G commands
		G = " ".join([x for x in CNC.vars["G"] if x[0]=="G"])	# remember $G
		TLO = CNC.vars["TLO"]
		self.softReset(False)			# reset controller
		self.control.purgeController()
		self.runEnded()
		self.stopProbe()
		if G: self.sendGCode(G)			# restore $G
		self.sendGCode("G43.1Z%s"%(TLO))	# restore TLO
		self.sendGCode("$G")

	#----------------------------------------------------------------------
	# Stop the current run
	#----------------------------------------------------------------------
	def stopRun(self, event=None):
		self.feedHold()
		self._stop = True
		# if we are in the process of submitting do not do anything
		if self._runLines != sys.maxint:
			self.purgeController()

	#----------------------------------------------------------------------
	# This should be called everytime that milling of g-code file is finished
	# So we can purge the controller for the next job
	# See https://github.com/vlachoudis/bCNC/issues/1035
	#----------------------------------------------------------------------
	def jobDone(self):
		print "Job done. Purging the controller"
		self.purgeController()

	#----------------------------------------------------------------------
	# This is called everytime that motion controller changes the state
	#----------------------------------------------------------------------
	def controllerStateChange(self, state):
		print "Controller state changed to: %s"%(state)

	#----------------------------------------------------------------------
	# thread performing I/O on serial line
	#----------------------------------------------------------------------
	def serialIO(self):
		self.sio_wait   = False		# wait for commands to complete (status change to Idle)
		self.sio_status = False		# waiting for status <...> report
		cline  = []		# length of pipeline commands
		sline  = []			# pipeline commands
		tosend = None			# next string to send
		tr = tg = time.time()		# last time a ? or $G was send to grbl

		while self.thread:
			t = time.time()
			# refresh machine position?
			if t-tr > SERIAL_POLL:
				self.serial.write(b"?")
				self.sio_status = True
				tr = t

				#If Override change, attach feed
				if CNC.vars["_OvChanged"]:
					self.control.overrideSet()

			# Fetch new command to send if...
			if tosend is None and not self.sio_wait and not self._pause and self.queue.qsize()>0:
				try:
					tosend = self.queue.get_nowait()
					#print "+++",repr(tosend)
					if isinstance(tosend, tuple):
						#print "gcount tuple=",self._gcount
						# wait to empty the grbl buffer and status is Idle
						if tosend[0] == WAIT:
							# Don't count WAIT until we are idle!
							self.sio_wait = True
							#print "+++ WAIT ON"
							#print "gcount=",self._gcount, self._runLines
						elif tosend[0] == MSG:
							# Count executed commands as well
							self._gcount += 1
							if tosend[1] is not None:
								# show our message on machine status
								self._msg = tosend[1]
						elif tosend[0] == UPDATE:
							# Count executed commands as well
							self._gcount += 1
							self._update = tosend[1]
						else:
							# Count executed commands as well
							self._gcount += 1
						tosend = None

					elif not isinstance(tosend,str) and not isinstance(tosend,unicode):
						try:
							tosend = self.gcode.evaluate(tosend)
#							if isinstance(tosend, list):
#								cline.append(len(tosend[0]))
#								sline.append(tosend[0])
							if isinstance(tosend,str) or isinstance(tosend,unicode):
								tosend += "\n"
							else:
								# Count executed commands as well
								self._gcount += 1
								#print "gcount str=",self._gcount
							#print "+++ eval=",repr(tosend),type(tosend)
						except:
							for s in str(sys.exc_info()[1]).splitlines():
								self.log.put((Sender.MSG_ERROR,s))
							self._gcount += 1
							tosend = None
				except Empty:
					break

				if tosend is not None:
					# All modification in tosend should be
					# done before adding it to cline
					if isinstance(tosend, unicode):
						tosend = tosend.encode("ascii","replace")

					# Keep track of last feed
					pat = FEEDPAT.match(tosend)
					if pat is not None:
						self._lastFeed = pat.group(2)

					# Modify sent g-code to reflect overrided feed for controllers without override support
					if not self.control.has_override:
						if CNC.vars["_OvChanged"]:
							CNC.vars["_OvChanged"] = False
							self._newFeed = float(self._lastFeed)*CNC.vars["_OvFeed"]/100.0
							if pat is None and self._newFeed!=0 \
							   and not tosend.startswith("$"):
								tosend = "f%g%s" % (self._newFeed, tosend)

						# Apply override Feed
						if CNC.vars["_OvFeed"] != 100 and self._newFeed != 0:
							pat = FEEDPAT.match(tosend)
							if pat is not None:
								try:
									tosend = "%sf%g%s\n" % \
										(pat.group(1),
										 self._newFeed,
										 pat.group(3))
								except:
									pass

					# Bookkeeping of the buffers
					sline.append(tosend)
					cline.append(len(tosend))

			# Anything to receive?
			if self.serial.inWaiting() or tosend is None:
				try:
					line = str(self.serial.readline()).strip()
				except:
					self.log.put((Sender.MSG_RECEIVE, str(sys.exc_info()[1])))
					self.emptyQueue()
					self.close()
					return

				#print "<R<",repr(line)
				#print "*-* stack=",sline,"sum=",sum(cline),"wait=",wait,"pause=",self._pause
				if not line:
					pass

				elif line[0]=="<":
					if not self.sio_status:
						self.log.put((Sender.MSG_RECEIVE, line))
					else:
						self.control.parseBracketAngle(line, cline)

				elif line[0]=="[":
					self.log.put((Sender.MSG_RECEIVE, line))
					self.control.parseBracketSquare(line)

				elif "error:" in line or "ALARM:" in line:
					self.log.put((Sender.MSG_ERROR, line))
					self._gcount += 1
					#print "gcount ERROR=",self._gcount
					if cline: del cline[0]
					if sline: CNC.vars["errline"] = sline.pop(0)
					if not self._alarm: self._posUpdate = True
					self._alarm = True
					CNC.vars["state"] = line
					if self.running:
						self._stop = True

				elif line.find("ok")>=0:
					self.log.put((Sender.MSG_OK, line))
					self._gcount += 1
					if cline: del cline[0]
					if sline: del sline[0]
					#print "SLINE:",sline
#					if  self._alarm and not self.running:
#						# turn off alarm for connected status once
#						# a valid gcode event occurs
#						self._alarm = False

				elif line[0] == "$":
					self.log.put((Sender.MSG_RECEIVE, line))
					pat = VARPAT.match(line)
					if pat:
						CNC.vars["grbl_%s"%(pat.group(1))] = pat.group(2)

				elif line[:4]=="Grbl" or line[:13]=="CarbideMotion": # and self.running:
					tg = time.time()
					self.log.put((Sender.MSG_RECEIVE, line))
					self._stop = True
					del cline[:]	# After reset clear the buffer counters
					del sline[:]
					CNC.vars["version"] = line.split()[1]
					# Detect controller
					if self.controller in ("GRBL0", "GRBL1"):
						self.controllerSet("GRBL%d"%(int(CNC.vars["version"][0])))

				else:
					self.log.put((Sender.MSG_RECEIVE, line))

			# Received external message to stop
			if self._stop:
				self.emptyQueue()
				tosend = None
				self.log.put((Sender.MSG_CLEAR, ""))
				# WARNING if runLines==maxint then it means we are
				# still preparing/sending lines from from bCNC.run(),
				# so don't stop
				if self._runLines != sys.maxint:
					self._stop = False

			#print "tosend='%s'"%(repr(tosend)),"stack=",sline,
			#	"sum=",sum(cline),"wait=",wait,"pause=",self._pause
			if tosend is not None and sum(cline) < RX_BUFFER_SIZE:
				self._sumcline = sum(cline)
#				if isinstance(tosend, list):
#					self.serial.write(str(tosend.pop(0)))
#					if not tosend: tosend = None

				#print ">S>",repr(tosend),"stack=",sline,"sum=",sum(cline)
				if self.control.gcode_case > 0: tosend = tosend.upper()
				if self.control.gcode_case < 0: tosend = tosend.lower()
				self.serial.write(bytes(tosend))
				#self.serial.write(tosend.encode("utf8"))
				#self.serial.flush()
				self.log.put((Sender.MSG_BUFFER,tosend))

				tosend = None
				if not self.running and t-tg > G_POLL:
					tosend = b"$G\n"
					sline.append(tosend)
					cline.append(len(tosend))
					tg = t



from Sender import ControllerGeneric, GPAT, STATUSPAT, POSPAT, TLOPAT, DOLLARPAT, FEEDPAT, SPLITPAT, VARPAT
from CNC import CNC
import time

class Controller(ControllerGeneric):
	def __init__(self, master):
		self.gcode_case = 1
		self.has_override = False
		self.master = master
		#print("smoothie loaded")

	def executeCommand(self, oline, line, cmd):
		if line[0] in ( "help", "version", "mem", "ls",
				"cd", "pwd", "cat", "rm", "mv",
				"remount", "play", "progress", "abort",
				"reset", "dfu", "break", "config-get",
				"config-set", "get", "set_temp", "get",
				"get", "net", "load", "save", "upload",
				"calc_thermistor", "thermistors", "md5sum"):
			self.master.serial.write(oline+"\n")
			return True
		return False

	def hardResetPre(self):
		self.master.serial.write(b"reset\n")

	def hardResetAfter(self):
		time.sleep(6)

	def viewSettings(self):
		pass

	def viewBuild(self):
		self.master.serial.write(b"version\n")
		self.master.sendGCode("$I")

	def viewStartup(self):
		pass

	def checkGcode(self):
		pass

	def grblHelp(self):
		self.master.serial.write(b"help\n")

	def grblRestoreSettings(self):
		pass

	def grblRestoreWCS(self):
		pass

	def grblRestoreAll(self):
		pass

	def purgeController(self):
		pass

	def overrideSet(self):
		pass

	def parseBracketAngle(self, line, cline):
		# <Idle|MPos:68.9980,-49.9240,40.0000,12.3456|WPos:68.9980,-49.9240,40.0000|F:12345.12|S:1.2>
		ln= line[1:-1] # strip off < .. >

		# split fields
		l= ln.split('|')

		# strip off status
		CNC.vars["state"]= l[0]

		# strip of rest into a dict of name: [values,...,]
		d= { a: [float(y) for y in b.split(',')] for a, b in [x.split(':') for x in l[1:]] }
		CNC.vars["mx"] = float(d['MPos'][0])
		CNC.vars["my"] = float(d['MPos'][1])
		CNC.vars["mz"] = float(d['MPos'][2])
		CNC.vars["wx"] = float(d['WPos'][0])
		CNC.vars["wy"] = float(d['WPos'][1])
		CNC.vars["wz"] = float(d['WPos'][2])
		CNC.vars["wcox"] = CNC.vars["mx"] - CNC.vars["wx"]
		CNC.vars["wcoy"] = CNC.vars["my"] - CNC.vars["wy"]
		CNC.vars["wcoz"] = CNC.vars["mz"] - CNC.vars["wz"]
		if 'F' in d:
		        CNC.vars["curfeed"] = float(d['F'][0])
		self.master._posUpdate = True

		# Machine is Idle buffer is empty
		# stop waiting and go on
		if self.master.sio_wait and not cline and l[0] in ("Idle","Check"):
		        self.master.sio_wait = False
		        self.master._gcount += 1

	def parseBracketSquare(self, line):
		pat = POSPAT.match(line)
		if pat:
			if pat.group(1) == "PRB":
				CNC.vars["prbx"] = float(pat.group(2))
				CNC.vars["prby"] = float(pat.group(3))
				CNC.vars["prbz"] = float(pat.group(4))
				#if self.running:
				self.master.gcode.probe.add(
					 CNC.vars["prbx"]
					+CNC.vars["wx"]
					-CNC.vars["mx"],
					 CNC.vars["prby"]
					+CNC.vars["wy"]
					-CNC.vars["my"],
					 CNC.vars["prbz"]
					+CNC.vars["wz"]
					-CNC.vars["mz"])
				self.master._probeUpdate = True
			CNC.vars[pat.group(1)] = \
				[float(pat.group(2)),
				 float(pat.group(3)),
				 float(pat.group(4))]
		else:
			pat = TLOPAT.match(line)
			if pat:
				CNC.vars[pat.group(1)] = pat.group(2)
				self.master._probeUpdate = True
			elif DOLLARPAT.match(line):
				CNC.vars["G"] = line[1:-1].split()
				CNC.updateG()
				self.master._gUpdate = True




from .stl import BUFFER_SIZE
from .stl import HEADER_SIZE
from .stl import COUNT_SIZE
from .stl import MAX_COUNT

from .stl import Mode
from .base import Dimension
from .base import RemoveDuplicates
from .mesh import Mesh

__all__ = [
    'BUFFER_SIZE',
    'HEADER_SIZE',
    'COUNT_SIZE',
    'MAX_COUNT',
    'Mode',
    'Dimension',
    'RemoveDuplicates',
    'Mesh',
]



#!/usr/bin/python
# -*- coding: ascii -*-

# Author: @harvie Tomas Mudrunka
# Date: 7 july 2018

__author__ = "@harvie Tomas Mudrunka"
#__email__  = ""

__name__ = _("Trochoidal")
__version__ = "0.0.2"

import math
import os.path
import re
from CNC import CNC,Block
from ToolsPage import Plugin
from math import pi, sqrt, sin, cos, asin, acos, atan2, hypot, degrees, radians, copysign, fmod

class Tool(Plugin):
	__doc__ = _("""Trochoidal g-code postprocessor""")			#<<< This comment will be show as tooltip for the ribbon button
	def __init__(self, master):
		Plugin.__init__(self, master,"Trochoidal")
		#Helical_Descent: is the name of the plugin show in the tool ribbon button
		self.icon = "helical"			#<<< This is the name of file used as icon for the ribbon button. It will be search in the "icons" subfolder
		self.group = "CAM"	#<<< This is the name of group that plugin belongs
		#Here we are creating the widgets presented to the user inside the plugin
		#Name, Type , Default value, Description
		self.variables = [			#<<< Define a list of components for the GUI
			("name"    ,    "db" ,    "", _("Name")),							#used to store plugin settings in the internal database
			("cw"    ,    "bool" ,    True, _("Clockwise")),
			("circ"    ,    "bool" ,    False, _("Circular")),
			("evenspacing"    ,    "bool" ,    True, _("Even spacing across segment")),
			("entry"    ,    "bool" ,    False, _("Trochoid entry (prepare for helicut)")),
			("rdoc"    ,    "mm" ,    "0.2", _("Radial depth of cut (<= cutter D * 0.4)")),
			("dia"    ,    "mm" ,    "3", _("Trochoid diameter (<= cutter D)")),
			("feed"    ,    "mm" ,    "2000", _("Feedrate"))
		]
		self.buttons.append("exe")  #<<< This is the button added at bottom to call the execute method below


	# ----------------------------------------------------------------------
	# This method is executed when user presses the plugin execute button
	# ----------------------------------------------------------------------
	def execute(self, app):
		feed = self["feed"]
		rdoc = self["rdoc"]
		radius = self["dia"]/2
		cw = self["cw"]
		circ = self["circ"]
		evenspacing = self["evenspacing"]

		if cw: cwtext = 'cw'
		else: cwtext = 'ccw'

		if cw: arcg = 'g2'
		else: arcg = 'g3'

		#print("go!")
		blocks  = []
		for bid in app.editor.getSelectedBlocks():
			#print(blocks[bid])
			path = app.gcode.toPath(bid)[0]
			#print(path)

			block = Block("trochoid "+cwtext+" "+str(radius*2)+"+"+str(rdoc))
			block.append("F"+str(feed))

			entry = self["entry"]

			A=path[0].A
			block.append("g0 x"+str(A[0])+" y"+str(A[1]))
			block.append("G1 Z0")
			for segment in path:
				#print(segment.A)
				#block.append("g0 x0 y0")
				#block.append("g1 x10 y10")
				#block.append("g1 x20 y10")
				#block.append("g0 x0 y0")
				if entry:
					eblock = Block("trochoid-in")
					eblock.append("G0 Z0")
					eblock.append("G0 x"+str(segment.A[0])+" y"+str(segment.A[1]-radius))
					eblock.append("G2 x"+str(segment.A[0])+" y"+str(segment.A[1]-radius)+" i"+str(0)+" j"+str(radius))
					blocks.append(eblock)
					entry = False

				#Continuity BEGINING
				block.append("g1 x"+str(segment.A[0])+" y"+str(segment.A[1]))
				#block.append(arcg+" x"+str(segment.A[0])+" y"+str(segment.A[1])+" r"+str(radius/2))

				phi = atan2(segment.B[1]-segment.A[1], segment.B[0]-segment.A[0])

				#TODO: handle arc segments
				#if segment.type == Segment.LINE:
				#if segment.type in (Segment.CW, Segment.CCW):

				#Compensate for uneven spacing
				srdoc = rdoc
				if evenspacing:
					subsegs = segment.length()//rdoc
					remainder = segment.length()%rdoc
					if remainder != 0:
						srdoc = segment.length()/(subsegs+1)

				#Loop over subsegmnents of segment
				i=0
				while i<(segment.length()+srdoc):
					pos=min(segment.length(), i)

					B = segment.distPoint(pos)
					block.extend(self.trochoid(A,B,radius,cw,circ))
					A = B

					i+=srdoc

				#Continuity END
				#block.append("g1 x"+str(segment.B[0])+" y"+str(segment.B[1]))
				block.append(arcg+" x"+str(segment.B[0])+" y"+str(segment.B[1])+" r"+str(radius/2))

			blocks.append(block)


		active = app.activeBlock()
		app.gcode.insBlocks(active, blocks, "Trochoidal created") #<<< insert blocks over active block in the editor
		app.refresh()                                                                                           #<<< refresh editor
		app.setStatus(_("Generated: Trochoidal"))                           #<<< feed back result
		#app.gcode.blocks.append(block)


	#Convert polar to cartesian and add that to existing vector
	def pol2car(self, r, phi, a=[0,0]):
		return [round(a[0]+r*cos(phi),4),round(a[1]+r*sin(phi),4)]

	#Generate single trochoidal element between two points
	def trochoid(self, A, B, radius, cw=True, circular=False):
		block = []

		if cw:
			u = 1
			arc = "G2"
		else:
			u = -1
			arc = "G3"


		phi = atan2(B[1]-A[1], B[0]-A[0])
		step = sqrt((A[0]-B[0])**2+(A[1]-B[1])**2)

		l = self.pol2car(radius, phi+radians(90*u))
		r = self.pol2car(radius, phi+radians(-90*u))
		al = self.pol2car(radius, phi+radians(90*u), A)
		ar = self.pol2car(radius, phi+radians(-90*u), A)
		bl = self.pol2car(radius, phi+radians(90*u), B)
		br = self.pol2car(radius, phi+radians(-90*u), B)

		# This schematic drawing represents naming convention
		# of points and vectors calculated in previous block
		#
		#    <--L---
		#          ---R-->
		#
		#        *   *
		#     *         *
		#    *           *
		#   BL     B     BR
		#    *           *
		#    *     ^     *
		#    *     |     *
		#    *     |     *
		#    *           *
		#   AL     A     AR
		#    *           *
		#     *         *
		#        *   *

		#TODO: improve strategies
		if circular:
			block.append("g1 x"+str(al[0])+" y"+str(al[1]))
			block.append("g1 x"+str(bl[0])+" y"+str(bl[1]))
			block.append(arc+" x"+str(bl[0])+" y"+str(bl[1])+" i"+str(r[0])+" j"+str(r[1]))
		else:
			#block.append(arc+" x"+str(al[0])+" y"+str(al[1])+" r"+str(radius/2))
			block.append("g1 x"+str(al[0])+" y"+str(al[1]))
			block.append("g1 x"+str(bl[0])+" y"+str(bl[1]))
			block.append(arc+" x"+str(br[0])+" y"+str(br[1])+" i"+str(r[0])+" j"+str(r[1]))
			block.append("g1 x"+str(ar[0])+" y"+str(ar[1]))
			block.append(arc+" x"+str(al[0])+" y"+str(al[1])+" i"+str(l[0])+" j"+str(l[1]))
			block.append("g1 x"+str(bl[0])+" y"+str(bl[1]))

		return block



#!/usr/bin/python
# -*- coding: ascii -*-
# $Id$
#
# Author: Mario Basz
#mariob_1960@yahoo.com.ar
# Date: 9 November 2017 
# Date: 03 may 2018
# A special thanks to Filippo Rivato and Vasilis. 
#This plugin is based on a variation
# of yours plugin Driller and My_Plugin example.
# To correct: Thats why the first point starts, 

__author__ = "Mario Basz"
__email__  = "mariob_1960@yahoo.com.ar"

__name__ = _("Trochoidal Path")
__version__= "0.0.5"

import math
from bmath import Vector
#import CNC					 # <<
from CNC import CNC,Block   #<< without this error it does not find CNC.vars
from ToolsPage import Plugin  

#==============================================================================
# Create Trochoidadl rute along selected blocks
#==============================================================================
class Tool(Plugin):
	__doc__ = _("Create a trochoid rute along selected blocks")
	def __init__(self, master):
		Plugin.__init__(self, master, "Trochoid_Path") #NAME OF THE PLUGIN
		self.icon  = "trochoidpath"
		self.group = "CAM"

		self.variables = [
			("name",      "db" ,    "", _("Name")),
	        ("CutDiam" ,"float", 10.0, _("Trochoid Diameter")),
			("Direction","inside,outside,on" , "outside", _("Direction")),
			("Offset",   "float",  0.0, _("Additional offset distance")),
			("endmill",   "db" ,    "", _("End Mill")),
			("Adaptative",  "bool",1,   _("Adaptative")),
			("Overcut",  "bool",     0, _("Overcut")),
			("cornerdiameter",  "float",    10, _("Diameter safe to corner %"))
		]
		self.buttons.append("exe")

	# ----------------------------------------------------------------------
	def execute(self, app):
		if self["endmill"]:
			self.master["endmill"].makeCurrent(self["endmill"])
		cornerdiameter = CNC.vars["diameter"]*(1+self["cornerdiameter"]/100.0)
		cornerradius = (self["CutDiam"] - cornerdiameter) /2.0
#		adaptedRadius = (self["CutDiam"] - CNC.vars["diameter"])/2.0

		direction = self["Direction"]
		name = self["name"]
		if name=="default" or name=="": name=None
#		app.trochprofile(direction, self["offset"], self["overcut"], name)
#		app.trochprofile(self["CutDiam"], direction, self["Offset"], self["Overcut"], self["Adaptative"], adaptedRadius, name)
		app.trochprofile(self["CutDiam"], direction, self["Offset"], self["Overcut"], self["Adaptative"], cornerradius, name)
		app.setStatus(_("Generate Trochoidal Profile path"))

#==============================================================================



#!/usr/bin/python
# -*- coding: ascii -*-

# Author: @harvie Tomas Mudrunka
# Date: 25 sept 2018

__author__ = "@harvie Tomas Mudrunka"
#__email__  = ""

__name__ = _("ArcFit")
__version__ = "0.1.1"

import math
import os.path
import re
from CNC import CNC,Block
from bpath import eq, Path, Segment
from ToolsPage import Plugin
from math import pi, sqrt, sin, cos, asin, acos, atan2, hypot, degrees, radians, copysign, fmod

class Tool(Plugin):
	__doc__ = _("""G-Code arc-fit""")			#<<< This comment will be show as tooltip for the ribbon button
	def __init__(self, master):
		Plugin.__init__(self, master,"ArcFit")
		self.icon = "arcfit"			#<<< This is the name of file used as icon for the ribbon button. It will be search in the "icons" subfolder
		self.group = "CAM"	#<<< This is the name of group that plugin belongs
		#self.oneshot = True
		#Here we are creating the widgets presented to the user inside the plugin
		#Name, Type , Default value, Description
		self.variables = [			#<<< Define a list of components for the GUI
			("name"    ,    "db" ,    "", _("Name")),							#used to store plugin settings in the internal database
			("preci", "mm", 0.5, _("arc precision (mm)"), _("how precisely must arc fit. set to 0 to disable arc fitting")),
			("linpreci", "mm", 0.001, _("line precision (mm)"), _("how precisely must line fit. set to 0 to disable line fitting, but at least some line fitting (0.001 to 0.01) might be needed to fix arcs, so they can be fit")),
			("numseg", "int", 3, _("minimal number of segments to create arc"))
		]
		self.buttons.append("exe")  #<<< This is the button added at bottom to call the execute method below
		self.help = """
This plugin will try to simplify the g-code by replacing amounts of short subsequent lines by one long arc.
There are some precision tunables which will allow you to define how much the resulting arc can differ from orignal lines.
This plugin can reverse the output of "linearize" plugin, which does the opposite.
This is not really meant to fillet sharp corners. But rather to reduce the number of g-code lines while preserving the toolpath shape.
This can be also useful for postprocessing of imported DXF/SVG splines. Splines have to be subdivided to short lines when importing and this can simplify the resulting code.
Another usecase is to postprocess mesh slices as STL/PLY format is based on triangles, it will never perfectly describe circles and arcs. You can use this plugin to simplify/smooth shapes imported from 3D mesh files.
Before this plugin tries to fit arcs it also tries to fit and merge longest possible lines within given precision. Line precision should be set much lower than arc precision, otherwise the line merging algorithm will "eat" lines that belong to arcs. Unless you want to do massive shape simplification and don't mind loosing details.
"""

	# ----------------------------------------------------------------------
	# This method is executed when user presses the plugin execute button
	# ----------------------------------------------------------------------
	def execute(self, app):
		preci = self.fromMm("preci")
		linpreci = self.fromMm("linpreci")
		numseg = self["numseg"]

		#print("go!")
		blocks  = []
		for bid in app.editor.getSelectedBlocks():
			if len(app.gcode.toPath(bid)) < 1: continue

			#nblock = Block("flat "+app.gcode[bid].name())
			#for i in app.gcode[bid]:
			#	nblock.append(re.sub(r"\s?z-?[0-9\.]+","",i))
			#blocks.append(nblock)

			eblock = Block("fit "+app.gcode[bid].name())
			npath = app.gcode.toPath(bid)[0]
			npath = npath.mergeLines(linpreci)
			npath = npath.arcFit(preci, numseg)
			if npath.length() <= 0:
				#FIXME: not sure how this could happen
				print "Warning: ignoring zero length path!"
				continue
			eblock = app.gcode.fromPath(npath,eblock)
			blocks.append(eblock)



		#active = app.activeBlock()
		#if active == 0: active+=1
		active=-1 #add to end
		app.gcode.insBlocks(active, blocks, "Arc fit") #<<< insert blocks over active block in the editor
		app.refresh()                                                                                           #<<< refresh editor
		app.setStatus(_("Generated: Arc fit"))                           #<<< feed back result
		#app.gcode.blocks.append(block)



#!/usr/bin/python
# -*- coding: ascii -*-
#
# Copyright and User License
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
# Copyright Vasilis.Vlachoudis@cern.ch for the
# European Organization for Nuclear Research (CERN)
#
# Please consult the flair documentation for the license
#
# DISCLAIMER
# ~~~~~~~~~~
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY, OF
# SATISFACTORY QUALITY, AND FITNESS FOR A PARTICULAR PURPOSE
# OR USE ARE DISCLAIMED. THE COPYRIGHT HOLDERS AND THE
# AUTHORS MAKE NO REPRESENTATION THAT THE SOFTWARE AND
# MODIFICATIONS THEREOF, WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADE SECRET OR OTHER PROPRIETARY RIGHT.
#
# LIMITATION OF LIABILITY
# ~~~~~~~~~~~~~~~~~~~~~~~
# THE COPYRIGHT HOLDERS AND THE AUTHORS SHALL HAVE NO
# LIABILITY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL,
# CONSEQUENTIAL, EXEMPLARY, OR PUNITIVE DAMAGES OF ANY
# CHARACTER INCLUDING, WITHOUT LIMITATION, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES, LOSS OF USE, DATA OR PROFITS,
# OR BUSINESS INTERRUPTION, HOWEVER CAUSED AND ON ANY THEORY
# OF CONTRACT, WARRANTY, TORT (INCLUDING NEGLIGENCE), PRODUCT
# LIABILITY OR OTHERWISE, ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGES.
#
# Author:	Vasilis.Vlachoudis@cern.ch
# Date:	29-Nov-2009

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import time
import fnmatch
from stat import *

if "_" not in dir(): _ = lambda x: x

try:
	from Tkinter import *
	import tkMessageBox as messagebox
except ImportError:
	from tkinter import *
	import tkinter.messagebox as messagebox
import tkExtra

_DIR_TYPE     = " <DIR>"
_FILE_TYPE    = "-file-"
_LINK_TYPE    = "-link-"
_BACKUP_TYPE  = "-backup-"
_TIME_FORMAT  = "%Y.%m.%d %H:%M:%S"
DISABLE_FILE  = "DarkGray"

COLORS = {
		"gz":	"Red",
		"tgz":	"Red",
		"zip":	"Red",
		"tbz":	"Red",
		"rpm":	"Red",
		"deb":	"Red",

		"flair":"DarkGreen",
		"fluka":"#109010",
		"inp":	"#109010",
		"out":	"#109010",
		"log":	"#109010",
		"err":	"Red",
		"lis":	"#109010",

		"py":	"Blue",
		"f":	"Blue",
		"F":	"Blue",
		"for":	"Blue",
		"FOR":	"Blue",
		"c":	"Blue",
		"C":	"Blue",
		"cc":	"Blue",
		"cpp":	"Blue",
		"a":	"Blue",
		"so":	"Blue",

		"eps":	"Purple",
		"ps":	"Purple",
		"gif":	"Purple",
		"png":	"Purple",
		"jpg":	"Purple",
		"bmp":	"Purple",
		"tif":	"Purple",

		"vxl":  "DarkRed",
		"dcm":  "DarkRed",

		"ngc":  "Brown",
		"nc" :  "Brown",

		"probe": "DarkOrange",
		"stl"  : "DarkOrange",

		_LINK_TYPE:	"DarkCyan",
		_BACKUP_TYPE:	"DarkGray",
		_DIR_TYPE:	"DarkBlue",
	}

DESCRIPTION = {
		"gz":	"Package gzip",
		"tgz":	"Package tgz",
		"zip":	"Package zip",
		"tbz":	"Package tbz",
		"rpm":	"Package rpm",
		"deb":	"Package deb",

		"flair":"FLAIR",
		"fluka":"Input",
		"inp":	"Input",
		"out":	"Output",
		"log":	"Log",
		"err":	"Error",
		"lis":	"Listing",

		"py":	"Python",
		"f":	"Fortran",
		"F":	"Fortran",
		"for":	"Fortran",
		"FOR":	"Fortran",
		"c":	"C",
		"C":	"C",
		"cc":	"C++",
		"cpp":	"C++",
		"a":	"Lib a",
		"so":	"Lib so",

		"eps":	"Image eps",
		"gif":	"Image gif",
		"jpg":	"Image jpg",
		"png":	"Image png",
		"bmp":	"Image bmp",
		"ps":	"Image ps",
		"tif":	"Image tif",

		"vxl":  "Voxel",
		"dcm":  "Dicom",

		_LINK_TYPE   : _LINK_TYPE,
		_BACKUP_TYPE : _BACKUP_TYPE,
		_DIR_TYPE    : _DIR_TYPE,
	}

# Converted from GIF to base64 PhotoImage
# base64.encodestring(open(iconfile,"rb").read())
_ICON = """
R0lGODlhEAAQAMZcABAQECEQECEYEDEgITEwIUI4MVJJQmNJIWNZQmNhWvw/Afw+DnNhUvxADv9B
D3NlSoRpIZRxIYRxUvRbM4R5Y6V5IfxhHvpnGvlsAZSCY7WCEPFwM5SKc5yKa5SKhPJ4QMaKENaK
EMaSAJyShPqDHqWShNaSAOuDZdaSEP+BSaWahP+LPqWilLWihLWilOWWfueiIeeiMbWqlP+ZauKg
itaqUveqMcaylOCql9ayc/eyMca6lN20o/e6QtbDhNbDlPfDQv+8i87Htf/DUs7LtefLhP/LUv/Q
Lv/RMf/TUufTpf/TY/fThP/Tc//bY/fbhP/bc//bhPfjlP/jhP/jlPfrpf3rlP/rpffzpffztf/z
pf31xf//////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////yH5
BAEKAH8ALAAAAAAQABAAAAe6gH+Cg4SFhoU4PIeFCjgKJy+KgjKUlEKCCgokDYMySllZWlcuRBYY
W1YXE4IuWKFXV1AsKitIRx+DLlWiV1FOTkEOKRsLNIIsVbA3N5QtLTMnJSUeHCpSV0UyU9vcU1BQ
NSPWVDI7UL9OS0lGPTEPfyVP2VNN6UtGQD06EB3wTDI/oCy5N0SfjRAHBHnwoe0ePh02YKCAkEAQ
Bxc5TGgUAUKDhgoRElrMQEECAwQGChAYIABAgEUwBQUCADs=
"""
_history     = []

#-------------------------------------------------------------------------------
def append2History(path):
	global _history
	if not path: return
	if path not in _history:
		_history.append(path)

#-------------------------------------------------------------------------------
def fileTypeColor(filename):
	fn = os.path.basename(filename)
	dot = fn.rfind(".")
	if dot>=0:
		ext = fn[dot+1:].lower()
	else:
		ext = _FILE_TYPE
	color  = None		# Default

	try: s = os.lstat(filename)
	except: return ext,color

	mode  = s[ST_MODE]
	isdir = S_ISDIR(mode)
	islnk = S_ISLNK(mode)

	if isdir:
		color = COLORS.get(_DIR_TYPE)
		ext   = DESCRIPTION.get(_DIR_TYPE)

	elif islnk:
		color = COLORS.get(_LINK_TYPE)
		ext   = DESCRIPTION.get(_LINK_TYPE)

	else:
		if fn[-1] == "~":
			color = COLORS.get(_BACKUP_TYPE)
		else:
			color = COLORS.get(ext,color)
			ext   = DESCRIPTION.get(ext,ext)

	return ext,color

#===============================================================================
# FileDialog
#===============================================================================
class FileDialog(Toplevel):
	_active   = False	# Avoid re-entrance of the dialog if by accident
				# someone double clicks a button
	sort      = None
	width     = -1
	height    = -1
	sash      = None
	newfolder = None

	# ----------------------------------------------------------------------
	def __init__(self, title=None,
			master=None,
			initialdir=None,
			initialfile=None,
			defaultextension=None,
			multiple=False,
			filetypes=None,
			**kw):

		Toplevel.__init__(self)
		self.transient(master)
		self.title(title)
		self.protocol("WM_DELETE_WINDOW", self.close)

		FileDialog._active = True

		if title is None: title = self._title

		self.dirframe = Frame(self)
		self.dirframe.pack(side=TOP, fill=X)
		Label(self.dirframe, text=_("Directory:")).grid(
			row=0, column=0)

		self.downButton = Label(self.dirframe, text=u"\u25BC",
				padx=3, pady=1, relief=RAISED)
		self.downButton.bind("<Button-1>", self.history)
		self.downButton.grid(row=0, column=99)

		if FileDialog.newfolder is None:
			FileDialog.newfolder = PhotoImage(data=_ICON)

		Button(self.dirframe, image=FileDialog.newfolder,
			padx=3, pady=3,
			command=self.newFolder).grid(
			row=0, column=100)
		self.dirframe.grid_columnconfigure(98,weight=1)
		self.buttons = []

		self.multiple = multiple
		if multiple:
			selectmode = EXTENDED
		else:
			selectmode = BROWSE
		self.fileList = tkExtra.ColorMultiListbox(self,
			((_("Name"), 30, None),
			 (_("Type"), 12, None),
			 (_("Size"),  8, None),
			 (_("Date"), 17, None)),
			 height=20,
			 selectmode=selectmode)
		self.fileList.pack(expand=YES, fill=BOTH)
		self.fileList.setPopupMenu(
			[('Rename', 0, self.rename),
			 ('Delete', 0, self.delete),
			 ('New Folder', 0, self.newFolder) ])

		self.fileList.bindList("<Double-1>",      self.double)
		self.fileList.bindList('<Return>',        self.double)
		self.fileList.bindList('<F2>',            self.rename)
		self.fileList.bindList("<Key-BackSpace>", self.upDir)
		self.fileList.bind("<<ListboxSelect>>",   self.select)
		self.fileList.bind("<<ListboxSort>>",     self.sortChanged)
		self.fileList.bind("<Configure>",         self.resize)

		frame = Frame(self)
		frame.pack(side=BOTTOM, fill=X)

		l = Label(frame, text=_("File name:"))
		l.grid(row=0, column=0, sticky=E)
		self.filename = Entry(frame, background="White")
		self.filename.grid(row=0, column=1, sticky=EW)
		self.ok = Button(frame, text=_("Open"), command=self.openFilename)
		self.ok.grid(row=0, column=2, sticky=EW)

		l = Label(frame, text=_("Files of type:"))
		l.grid(row=1, column=0, sticky=E)

		self.typeCombo = tkExtra.Combobox(frame, command=self.typeChange)
		self.typeCombo.grid(row=1, column=1, sticky=NSEW)

		self.filter = None
		self.filetypes = {}
		if filetypes:
			if isinstance(filetypes[0],str):
				filetypes = (filetypes,)
			self.filter = filetypes[0]
			for desc,ext in filetypes:
				if isinstance(ext,str):
					s = "%s (%s)"%(desc,ext)
					ext = (ext,)
				else:
					s = "%s (%s)"%(desc, ",".join(ext))
				self.typeCombo.insert(END, s)
				self.filetypes[s] = ext

		b = Button(frame, text=_("Cancel"), command=self.cancel)
		b.grid(row=1, column=2, sticky=EW)
		frame.grid_columnconfigure(1, weight=1)

		self.bind("<Return>", self.openFilename)
		self.bind("<Escape>", self.cancel)

		# Variables
		# 1st set the path if any
		if initialdir:
			self.path = os.path.abspath(initialdir)
		else:	# otherwise to the current directory
			self.path = os.getcwd()

		if initialfile:
			# if a file is specified and has a different path
			initialfile = os.path.abspath(initialfile)
			d,n = os.path.split(initialfile)
			if d != "": self.path = d
			self.filename.insert(0, n)
			self.filename.select_range(0,END)

		# Flags
		self.hidden = False	# Show hidden files
		self.links  = True	# Show links
		self.dirs   = True	# Show directories
		self.files  = True	# Show files
		self.seldir = False	# Select directory instead of file
		self.selFile = ""	# Selected files
		append2History(self.path)

		# popup history
		self._popup = None
		self._historyOldLen = len(_history)

	# ----------------------------------------------------------------------
	def show(self):
		self.deiconify()
		if FileDialog.width > 0:
			self.geometry("%dx%d" \
				%(FileDialog.width, FileDialog.height))

		self.buttonPath(self.path)
		self.typeCombo.set(self.typeCombo.get(0))	# will fill the files
		try:
			self.lift()
			self.focus_set()
			self.filename.focus_set()
			self.wait_visibility()
			self.grab_set()
			self.wait_window()
		except TclError:
			pass
		return self.selFile

	# ----------------------------------------------------------------------
	# Restore sash size on creation
	# ----------------------------------------------------------------------
	def resize(self, event):
		if FileDialog.sash is not None:
			FileDialog.sash.reverse()
			# XXX XXX ERROR: paneframe doesn't update the width/height
			# so all sash placement is wrong
			self.fileList.paneframe.update()
			#self.fileList.paneframe.update_idletasks()
			if FileDialog.sash is not None:
				n = len(FileDialog.sash)-1
				for i,x in enumerate(FileDialog.sash):
					self.fileList.paneframe.sash_place(n-i,x,1)
				FileDialog.sash = None

	# ----------------------------------------------------------------------
	# Create buttons for the path
	# ----------------------------------------------------------------------
	def buttonPath(self, path):
		path = path.split(os.sep)
		if path[0] == "": path[0] = os.sep
		if path[-1] == "": del path[-1]
		lp = len(path)
		lb = len(self.buttons)

		i = 0
		same = True
		while i < min(lp, lb):
			b = self.buttons[i]
			if path[i] != b["text"]:
				b["text"] = path[i]
				same = False
			i += 1

		if lb < lp:	# Create new buttons
			while i < lp:
				self.buttons.append(Button(self.dirframe,
					text=path[i],
					command=lambda s=self,b=i:s.button(b),
					padx=1))
				i += 1
				self.buttons[-1].grid(row=0, column=i)

		elif lp < lb and not same:	# Use existing buttons
			while i < lb:
				self.buttons[i].grid_forget()
				i += 1
			del self.buttons[lp:]

		for i in range(lp):
			self.buttons[i]["foreground"] = "black"
			self.buttons[i]["activeforeground"] = "black"
		self.buttons[lp-1]["foreground"] = "blue"
		self.buttons[lp-1]["activeforeground"] = "blue"
		for i in range(lp, len(self.buttons)):
			self.buttons[i]["foreground"] = "darkgray"
			self.buttons[i]["activeforeground"] = "darkgray"

	# ----------------------------------------------------------------------
	def button(self, b):
		path = [ x["text"] for x in self.buttons[0:b+1] ]
		if path[0] == os.sep:
			path = os.sep + os.sep.join(path[1:])
		else:
			path = os.sep.join(path)
		if path=="": path=os.sep
		self.fileList.focus_set()
		self.changePath(path)

	# ----------------------------------------------------------------------
	def history(self, event=None):
		if self._popup is not None:
			self._historyDestroy()
			return
		self.downButton["relief"] = SUNKEN
		self._popup = Toplevel(self)
		self._popup.transient(self)
		self._popup.overrideredirect(1)
		self._popup.withdraw()
		self._popup.bind('<Escape>',	self._historyDestroy)
		self._popup.bind('<FocusOut>',	self._historyFocusOut)

		x = self.buttons[0].winfo_rootx()
		y = self.buttons[0].winfo_rooty() + self.buttons[0].winfo_height() - 2
		w = self.downButton.winfo_rootx() + self.downButton.winfo_width() - x
		h = self.fileList.winfo_height() - 20
		self._popup.geometry('%dx%d+%d+%d' % (w,h,x,y))

		sb = Scrollbar(self._popup, orient=VERTICAL, takefocus=False)
		sb.pack(side=RIGHT, fill=Y)
		self._popupList = Listbox(self._popup,
			background="White",
			selectmode=BROWSE,
			takefocus=True,
			yscrollcommand=sb.set)
		self._popupList.pack(side=LEFT, fill=BOTH, expand=YES)
		self._popupList.bind("<ButtonRelease-1>", self._historyClick)
		self._popupList.bind("<Return>", self._historyClick)
		sb.config(command=self._popupList.yview)
		for h in sorted(_history):
			self._popupList.insert(END, h)
		self._popupList.selection_set(0)
		self._popupList.activate(0)

		self.grab_release()
		self._popup.deiconify()
		self._popup.lift()
		self._popup.focus_set()
		self._popupList.focus_set()
		self._popup.update_idletasks()

	# ----------------------------------------------------------------------
	def _historyFocusOut(self, event):
		try:
			f = self.focus_get()
		except KeyError:
			pass
		else:
			if f == self._popup or f == self._popupList:
				return
		self._historyDestroy()

	# ----------------------------------------------------------------------
	def _historyClick(self, event):
		try:
			sel = self._popupList.curselection()[0]
			self.changePath(self._popupList.get(sel))
			self._historyDestroy()
		except:
			pass

	# ----------------------------------------------------------------------
	def _historyDestroy(self, event=None):
		self.downButton["relief"] = RAISED
		self._popup.destroy()
		self._popup = None
		self.grab_set()
		self.focus_set()
		return "break"

	# ----------------------------------------------------------------------
	def upDir(self, event):
		if not tkExtra.ExListbox._search:
			self.changePath(os.path.dirname(self.path))
		else:
			event.widget.handleKey(event)

	# ----------------------------------------------------------------------
	def changePath(self, path):
		if path[-1] != os.sep: path += os.sep
		path = os.path.abspath(path)
		try: os.lstat(path)
		except OSError:
			messagebox.showerror(_("Error"),
				_("Cannot access path \"%s\"")%(path),
				parent=self)
			return
		self.buttonPath(path)
		self.path = path
		self.fill()
		append2History(self.path)

	# ----------------------------------------------------------------------
	def fill(self, path=None):
		self.fileList.delete(0,END)
		self.fileList.listbox(0).resetSearch()

		if path is None: path = self.path

		# Populate list but sorted
		try:
			for fn in os.listdir(path):
				if not self.hidden and fn[0]==".": continue
				filename = os.path.join(path, fn)
				ext, color = fileTypeColor(filename)

				try: s = os.lstat(filename)
				except: continue

				size  = 0
				islnk = S_ISLNK(s[ST_MODE])
				if islnk:
					try: s = os.stat(filename)
					except: continue
				isdir = S_ISDIR(s[ST_MODE])

				if self.filter is not None and not isdir and not islnk:
					match = False
					for pat in self.filter:
						if fnmatch.fnmatch(fn, pat):
							match = True
							break
				else:
					match = True

				if isdir:
					if not self.dirs: continue
				elif islnk:
					if not self.links: continue
				else:
					size = s[ST_SIZE]

				if match:
					self.fileList.insert(END, (fn, ext, size,
						 time.strftime(_TIME_FORMAT,
							time.localtime(s[ST_MTIME]))))
					if not self.files and not isdir:
						self.fileList.setColor(END, DISABLE_FILE)
					elif color:
						self.fileList.setColor(END, color)
		except OSError:
			messagebox.showerror(_("Error"),
				_("Error listing folder \"%s\"")%(path),
				parent=self)

		if FileDialog.sort is None:
			self.fileList.sort(0, False)		# First short by name
			# Move all directories to top
			self.fileList.sort(1, False)		# then by type
			FileDialog.sort = None
		else:
			self.fileList.restoreSort(FileDialog.sort)

		# Find item to select
		fn = self.filename.get()
		if self.seldir: fn = os.path.basename(fn)
		for i in range(self.fileList.size()):
			if fn == self.fileList.listbox(0).get(i):
				self.fileList.see(i)
				self.fileList.activate(i)
				self.fileList.selection_set(i)
				break
		else:
			self.fileList.see(0)
			self.fileList.activate(0)
			self.fileList.selection_set(0)

	# ----------------------------------------------------------------------
	def sortChanged(self, event=None):
		FileDialog.sort = self.fileList.saveSort()

	# ----------------------------------------------------------------------
	def open(self, fn):
		if self.seldir and self.path == fn:
			self.selFile = self.path

		# Single file selection
		elif fn.find('","')<0:
			# Check for path
			try:
				filename = os.path.join(self.path, fn)
				s = os.stat(filename)
				if S_ISDIR(s[ST_MODE]):
					self.changePath(filename)
					self.filename.delete(0,END)
					return
			except OSError:
				pass

			# Check for a pattern
			if fn.find('*')>=0 or fn.find('?')>=0:
				self.filter = (fn,)
				self.fill()
				return

			# Check for extension
			if self.filter:
				fn,ext = os.path.splitext(filename)
				if ext == "":
					ffn,ffext = os.path.splitext(self.filter[0])
					if ffext!="":
						filename = fn+ffext

			if self.multiple:
				self.selFile = [filename]
			else:
				self.selFile = filename

		# Multiple file selection
		else:
			self.selFile = [os.path.join(self.path,f) \
					for f in fn[1:-1].split('","')]

		if self.check():
			global _history
			# Delete all temporary directories and keep only the last one
			if len(_history) > self._historyOldLen:
				del _history[self._historyOldLen:]
			append2History(self.path)
			self.close()

	# ----------------------------------------------------------------------
	# Open the filename entered in the entry box
	# ----------------------------------------------------------------------
	def openFilename(self, event=None):
		fn = self.filename.get()
		# Single file selection?
		if fn=="":
			self.select()
			fn = self.filename.get()
			if fn=="": return
		self.open(fn)

	# ----------------------------------------------------------------------
	def check(self): return True

	# ----------------------------------------------------------------------
	def cancel(self, event=None):
		if self._popup is not None:
			self._historyDestroy()
		else:
			self.selFile = ""
			self.close()

	# ----------------------------------------------------------------------
	def close(self):
		FileDialog._active = False
		FileDialog.width   = self.winfo_width()
		FileDialog.height  = self.winfo_height()
		FileDialog.sash    = [self.fileList.paneframe.sash_coord(i)[0]
					for i in range(len(self.fileList.listboxes())-1)]
		tkExtra.ExListbox.resetSearch()
		self.grab_release()
		self.destroy()

	# ----------------------------------------------------------------------
	def double(self, event):
		sel = self.fileList.curselection()
		if len(sel)!=1: return
		item = self.fileList.get(sel[0])
		if item[1] == _DIR_TYPE:
			self.changePath(os.path.join(self.path, item[0]))
			return "break"
		elif item[1] == _LINK_TYPE:
			# maybe a directory?
			path = os.path.join(self.path, item[0])
			try:
				s = os.stat(path)
				if S_ISDIR(s[ST_MODE]):
					self.changePath(path)
					return "break"
			except: pass
		self.openFilename()

	# ----------------------------------------------------------------------
	# Select current file from listbox
	# ----------------------------------------------------------------------
	def select(self, event=None):
		sel = self.fileList.curselection()

		if len(sel)==1:
			item = self.fileList.get(sel[0])[0]
			fn = os.path.join(self.path, item)
			if self.seldir:
				try:
					s = os.stat(fn)
					if not S_ISDIR(s[ST_MODE]):
						fn = os.path.dirname(fn)
				except OSError:
					pass
				self.filename.delete(0, END)
				self.filename.insert(0, fn)
			else:
				try:
					s = os.stat(fn)
					if not S_ISDIR(s[ST_MODE]):
						self.filename.delete(0, END)
						self.filename.insert(0, item)
				except OSError:
					pass
		else:
			lget = self.fileList.get
			files = ["\"%s\""%(lget(i)[0]) for i in sel]

			if files:
				self.filename.delete(0, END)
				self.filename.insert(0, ",".join(files))

			elif self.seldir:
				self.filename.delete(0, END)
				self.filename.insert(0, self.path)

	# ----------------------------------------------------------------------
	def typeChange(self, event=None):
		pat = self.typeCombo.get()
		self.filter = self.filetypes.get(pat,None)
		self.fill()
		if self.filter is None or self.seldir: return

		# Change extension if needed
		first = None
		filename = self.filename.get()
		if filename == "" or "," in filename: return
		fn,ext = os.path.splitext(filename)
		for i in self.filter:
			f,e = os.path.splitext(i)
			if first is None and e: first = e
			if e == ext: return
		else:
			if first:
				# not found, change the filename to the first extension
				self.filename.delete(0, END)
				self.filename.insert(0, fn+first)

	# ----------------------------------------------------------------------
	def newFolder(self):
		self.fileList.insert(END, (_("NewFolder"), _DIR_TYPE, 0,
					 time.strftime(_TIME_FORMAT,
					 time.localtime(time.time()))))
		self.fileList.see(END)
		self.fileList.selection_clear(0,END)
		self.fileList.selection_set(END)
		self.fileList.activate(END)
		edit = tkExtra.InPlaceEdit(self.fileList.listbox(0))
		if edit.value:
			try:
				os.mkdir(os.path.join(self.path, edit.value))
			except OSError:
				messagebox.showerror(_("Error"),
					_("Error creating folder \"%s\"")%(edit.value),
					parent=self)
				self.fileList.delete(END)
				return
			self.fileList.selection_set(END)
			self.fileList.see(END)
			self.fileList.setColor(END, COLORS.get(_DIR_TYPE))
			self.select()
		else:
			try:
				self.fileList.delete(END)
			except TclError:
				pass

	# ----------------------------------------------------------------------
	def rename(self,event=None):
		fn = self.fileList.listbox(0).get(ACTIVE)
		edit = tkExtra.InPlaceEdit(self.fileList.listbox(0))
		if edit.value and edit.value != fn:
			try:
				os.rename(os.path.join(self.path, fn),
					os.path.join(self.path, edit.value))
			except OSError:
				messagebox.showerror(_("Error"),
					_("Error renaming \"%s\" to \"%s\"") \
						%(fn, edit.value),
					parent=self)
		self.select()

	# ----------------------------------------------------------------------
	def delete(self):
		sel = map(int,self.fileList.curselection())
		sel.reverse()
		if not sel: return
		try:
			for i in sel:
				fn = self.fileList.listbox(0).get(i)
				filename = os.path.join(self.path, fn)
				s = os.lstat(filename)
				if S_ISDIR(s[ST_MODE]):
					os.rmdir(filename)
				else:
					os.remove(filename)
				self.fileList.delete(i)
		except OSError:
			messagebox.showerror(_("Error"),
					_("Error deleting file \"%s\"")%(fn),
					parent=self)
		self.select()

#===============================================================================
class OpenDialog(FileDialog):
	_title = _("Open")

	# ----------------------------------------------------------------------
	# Check if file exist
	# ----------------------------------------------------------------------
	def check(self):
		if isinstance(self.selFile, list):
			for f in self.selFile:
				try:
					os.lstat(f)
				except:
					messagebox.showwarning(_("File does not exist"),
						_("File \"%s\" does not exist")%(f),
						parent=self)
					self.selFile = ""
					return False
		else:
			try:
				os.lstat(self.selFile)
			except:
				messagebox.showwarning(_("File does not exist"),
					_("File \"%s\" does not exist")%(self.selFile),
					parent=self)
				self.selFile = ""
				return False
		return True

#===============================================================================
class SaveAsDialog(FileDialog):
	_title = _("Save As")

	# ----------------------------------------------------------------------
	def __init__(self, **kw):
		FileDialog.__init__(self, **kw)
		self.ok["text"] = _("Save")

	# ----------------------------------------------------------------------
	def check(self):
		try:
			os.lstat(self.selFile)
			ans = messagebox.askyesno(_("File already exists"),
				_("Overwrite existing file %r?")%(self.selFile),
				parent=self)
			if str(ans)!=messagebox.YES and not ans:
				self.selFile = ""
				return False
		except:
			pass
		return True

#===============================================================================
class DirectoryDialog(FileDialog):
	_title = _("Choose Directory")
	def __init__(self, **kw):
		FileDialog.__init__(self, **kw)
		self.files  = False
		self.seldir = True
		self.filename.insert(0, self.path)

	# ----------------------------------------------------------------------
	def changePath(self, path):
		path = os.path.abspath(path)
		FileDialog.changePath(self, path)
		self.filename.delete(0, END)
		self.filename.insert(0, path)

#===============================================================================
def askfilename(**options):
	"""Ask for a filename"""
	if FileDialog._active: return ""
	return FileDialog(**options).show()

def askopenfilename(**options):
	"""Ask for a filename to open"""
	if FileDialog._active: return ""
	return OpenDialog(**options).show()

def askopenfilenames(**options):
	"""Ask for a multiple filenames to open"""
	if FileDialog._active: return ()
	options["multiple"] = True
	return OpenDialog(**options).show()

def asksaveasfilename(**options):
	"""Ask for a filename to save as"""
	if FileDialog._active: return ""
	return SaveAsDialog(**options).show()

def askdirectory(**options):
	"""Ask for a directory"""
	if FileDialog._active: return ""
	return DirectoryDialog(**options).show()

#===============================================================================
if __name__ == "__main__":
	import sys

	root = Tk()
	root.withdraw()
	initdir = None
	if len(sys.argv)>1:
		initdir = os.path.abspath(sys.argv[1])
	#print askdirectory()

	files = asksaveasfilename(title=_("Open"),
			initialdir=initdir,
#			initialfile="test.f",
			filetypes=(("All","*"),
				("Python", "*.py"),
				("Flair",("*.flair", "*.inp"))))
	#print files

	#import tkFileDialog
	#print tkFileDialog.asksaveasfilename( title="Open flair project",
	#		initialdir=initdir,
	#		filetypes=(("All","*"),
	#			("Python", "*.py"),
	#			("Flair",("*.flair", "*.inp"))))



"""
Functions to slice a mesh. For now, computes planar cross-section
"""
import numpy as np
import numpy.linalg as la
try:
    import scipy.spatial.distance as spdist
    USE_SCIPY = True
except ImportError:
    USE_SCIPY = False
import collections

# ---- Geometry datastructures


def make_edge(v1, v2):
    """
    We store edges as tuple where the vertex indices are sorted (so
    the edge going from v1 to v2 and v2 to v1 is the same)
    """
    return tuple(sorted((v1, v2)))


class TriangleMesh(object):
    def __init__(self, verts, tris):
        """
        Args:
            verts: The 3D vertex positions
            tris: A list of triplet containing vertex indices for each triangle
        """
        self.verts = np.array(verts)
        # For each edge, contains the list of triangles it belongs to
        # If the mesh is closed, each edge belongs to 2 triangles
        self.edges_to_tris = collections.defaultdict(lambda: [])
        # For each triangle, contains the edges it contains
        self.tris_to_edges = {}
        # For each vertex, the list of triangles it belongs to
        self.verts_to_tris = collections.defaultdict(lambda: [])

        self.tris = tris

        # Fill data structures
        for tid, f in enumerate(tris):
            tri_edges = []
            for i in range(3):
                v1 = f[i]
                v2 = f[(i + 1) % 3]
                e = make_edge(v1, v2)
                self.edges_to_tris[e].append(tid)
                tri_edges.append(e)
                self.verts_to_tris[f[i]].append(tid)
            self.tris_to_edges[tid] = tri_edges

        # Sanity check : max 2 faces per edge
        for e, tris in self.edges_to_tris.items():
            assert len(tris) <= 2

    def edges_for_triangle(self, tidx):
        """Returns the edges forming triangle with given index"""
        return self.tris_to_edges[tidx]

    def triangles_for_edge(self, edge):
        return self.edges_to_tris[edge]

    def triangles_for_vert(self, vidx):
        """Returns the triangles `vidx` belongs to"""
        return self.verts_to_tris[vidx]


class Plane(object):
    def __init__(self, orig, normal):
        self.orig = orig
        self.n = normal / la.norm(normal)

    def __str__(self):
        return 'plane(o=%s, n=%s)' % (self.orig, self.n)


def point_to_plane_dist(p, plane):
    return np.dot((p - plane.orig), plane.n)


def triangle_intersects_plane(mesh, tid, plane):
    """
    Returns true if the given triangle is cut by the plane. This will return
    false if a single vertex of the triangle lies on the plane
    """
    dists = [point_to_plane_dist(mesh.verts[vid], plane)
             for vid in mesh.tris[tid]]
    side = np.sign(dists)
    return not (side[0] == side[1] == side[2])


# ---- Planar cross-section

INTERSECT_EDGE = 0
INTERSECT_VERTEX = 1


def compute_triangle_plane_intersections(mesh, tid, plane, dist_tol=1e-8):
    """
    Compute the intersection between a triangle and a plane

    Returns a list of intersections in the form
        (INTERSECT_EDGE, <intersection point>, <edge>) for edges intersection
        (INTERSECT_VERTEX, <intersection point>, <vertex index>) for vertices


    This return between 0 and 2 intersections :
    - 0 : the plane does not intersect the plane
    - 1 : one of the triangle's vertices lies on the plane (so it just
          "touches" the plane without really intersecting)
    - 2 : the plane slice the triangle in two parts (either vertex-edge,
          vertex-vertex or edge-edge)
    """
    # TODO: Use a distance cache
    dists = {vid: point_to_plane_dist(mesh.verts[vid], plane)
             for vid in mesh.tris[tid]}
    # TODO: Use an edge intersection cache (we currently compute each edge
    # intersection twice : once for each tri)

    # This is to avoid registering the same vertex intersection twice
    # from two different edges
    vert_intersect = {vid: False for vid in dists.keys()}

    # Iterate through the edges, cutting the ones that intersect
    intersections = []
    for e in mesh.edges_for_triangle(tid):
        v1 = mesh.verts[e[0]]
        d1 = dists[e[0]]
        v2 = mesh.verts[e[1]]
        d2 = dists[e[1]]

        if np.fabs(d1) < dist_tol:
            # Avoid creating the vertex intersection twice
            if not vert_intersect[e[0]]:
                # point on plane
                intersections.append((INTERSECT_VERTEX, v1, e[0]))
                vert_intersect[e[0]] = True
        if np.fabs(d2) < dist_tol:
            if not vert_intersect[e[1]]:
                # point on plane
                intersections.append((INTERSECT_VERTEX, v2, e[1]))
                vert_intersect[e[1]] = True

        # If vertices are on opposite sides of the plane, we have an edge
        # intersection
        if d1 * d2 < 0:
            # Due to numerical accuracy, we could have both a vertex intersect
            # and an edge intersect on the same vertex, which is impossible
            if not vert_intersect[e[0]] and not vert_intersect[e[1]]:
                # intersection factor (between 0 and 1)
                # here is a nice drawing :
                # https://ravehgonen.files.wordpress.com/2013/02/slide8.png
                # keep in mind d1, d2 are *signed* distances (=> d1 - d2)
                s = d1 / (d1 - d2)
                vdir = v2 - v1
                ipos = v1 + vdir * s
                intersections.append((INTERSECT_EDGE, ipos, e))

    return intersections


def get_next_triangle(mesh, T, plane, intersection, dist_tol):
    """
    Returns the next triangle to visit given the intersection and
    the list of unvisited triangles (T)

    We look for a triangle that is cut by the plane (2 intersections) as
    opposed to one that only touch the plane (1 vertex intersection)
    """
    if intersection[0] == INTERSECT_EDGE:
        tris = mesh.triangles_for_edge(intersection[2])
    elif intersection[0] == INTERSECT_VERTEX:
        tris = mesh.triangles_for_vert(intersection[2])
    else:
        assert False, 'Invalid intersection[0] value : %d' % intersection[0]

    # Knowing where we come from is not enough. If an edge of the triangle
    # lies exactly on the plane, i.e. :
    #
    #   /t1\
    # -v1---v2-
    #   \t2/
    #
    # With v1, v2 being the vertices and t1, t2 being the triangles, then
    # if you just try to go to the next connected triangle that intersect,
    # you can visit v1 -> t1 -> v2 -> t2 -> v1 .
    # Therefore, we need to limit the new candidates to the set of unvisited
    # triangles and once we've visited a triangle and decided on a next one,
    # remove all the neighbors of the visited triangle so we don't come
    # back to it

    T = set(T)
    for tid in tris:
        if tid in T:
            intersections = compute_triangle_plane_intersections(
                    mesh, tid, plane, dist_tol)
            if len(intersections) == 2:
                T = T.difference(tris)
                return tid, intersections, T
    return None, [], T


def _walk_polyline(tid, intersect, T, mesh, plane, dist_tol):
    """
    Given an intersection, walk through the mesh triangles, computing
    intersection with the cut plane for each visited triangle and adding
    those intersection to a polyline.
    """
    T = set(T)
    p = []
    # Loop until we have explored all the triangles for the current
    # polyline
    while True:
        p.append(intersect[1])

        tid, intersections, T = get_next_triangle(mesh, T, plane,
                                                  intersect, dist_tol)
        if tid is None:
            break

        # get_next_triangle returns triangles that our plane actually
        # intersects (as opposed to touching only a single vertex),
        # hence the assert
        assert len(intersections) == 2

        # Of the two returned intersections, one should have the
        # intersection point equal to p[-1]
        if la.norm(intersections[0][1] - p[-1]) < dist_tol:
            intersect = intersections[1]
        else:
            assert la.norm(intersections[1][1] - p[-1]) < dist_tol, \
                '%s not close to %s' % (str(p[-1]), str(intersections))
            intersect = intersections[0]

    return p, T


def cross_section_mesh(mesh, plane, dist_tol=1e-8):
    """
    Args:
        mesh: A geom.TriangleMesh instance
        plane: The cut plane : geom.Plane instance
        dist_tol: If two points are closer than dist_tol, they are considered
                  the same
    """
    # Set of all triangles
    T = set(range(len(mesh.tris)))
    # List of all cross-section polylines
    P = []

    while len(T) > 0:
        tid = T.pop()

        intersections = compute_triangle_plane_intersections(
                mesh, tid, plane, dist_tol)

        if len(intersections) == 2:
            for intersection in intersections:
                p, T = _walk_polyline(tid, intersection, T, mesh, plane,
                                      dist_tol)
                if len(p) > 1:
                    P.append(np.array(p))
    return P


def cross_section(verts, tris, plane_orig, plane_normal, **kwargs):
    """
    Compute the planar cross section of a mesh. This returns a set of
    polylines.

    Args:
        verts: Nx3 array of the vertices position
        faces: Nx3 array of the faces, containing vertex indices
        plane_orig: 3-vector indicating the plane origin
        plane_normal: 3-vector indicating the plane normal

    Returns:
        A list of Nx3 arrays, each representing a disconnected portion
        of the cross section as a polyline
    """
    mesh = TriangleMesh(verts, tris)
    plane = Plane(plane_orig, plane_normal)
    return cross_section_mesh(mesh, plane, **kwargs)


def pdist_squareformed_numpy(a):
    """
    Compute spatial distance using pure numpy
    (similar to scipy.spatial.distance.cdist())

    Thanks to Divakar Roy (@droyed) at stackoverflow.com

    Note this needs at least np.float64 precision!

    Returns: dist
    """
    a = np.array(a, dtype=np.float64)
    a_sumrows = np.einsum('ij,ij->i', a, a)
    dist = a_sumrows[:, None] + a_sumrows - 2 * np.dot(a, a.T)
    np.fill_diagonal(dist, 0)
    return dist


def merge_close_vertices(verts, faces, close_epsilon=1e-5):
    """
    Will merge vertices that are closer than close_epsilon.

    Warning, this has a O(n^2) memory usage because we compute the full
    vert-to-vert distance matrix. If you have a large mesh, might want
    to use some kind of spatial search structure like an octree or some fancy
    hashing scheme

    Returns: new_verts, new_faces
    """
    # Pairwise distance between verts
    if USE_SCIPY:
        D = spdist.cdist(verts, verts)
    else:
        D = np.sqrt(np.abs(pdist_squareformed_numpy(verts)))

    # Compute a mapping from old to new : for each input vert, store the index
    # of the new vert it will be merged into
    old2new = np.zeros(D.shape[0], dtype=np.int)
    # A mask indicating if a vertex has already been merged into another
    merged_verts = np.zeros(D.shape[0], dtype=np.bool)
    new_verts = []
    for i in range(D.shape[0]):
        if merged_verts[i]:
            continue
        else:
            # The vertices that will be merged into this one
            merged = np.flatnonzero(D[i, :] < close_epsilon)
            old2new[merged] = len(new_verts)
            new_verts.append(verts[i])
            merged_verts[merged] = True

    new_verts = np.array(new_verts)

    # Recompute face indices to index in new_verts
    new_faces = np.zeros((len(faces), 3), dtype=np.int)
    for i, f in enumerate(faces):
        new_faces[i] = (old2new[f[0]], old2new[f[1]], old2new[f[2]])

    # again, plot with utils.trimesh3d(new_verts, new_faces)
    return new_verts, new_faces
